from .bigg import create_bigg_universal_model, get_bigg_model_list, create_cobra_model_from_bigg_model, \
    get_bigg_metabolite, add_bigg_metabolites, get_bigg_reaction, add_bigg_reactions
from .metanetx import create_metanetx_universal_model
from .source import create_universal_model_from_source, load_model_from_file
from .vmh import create_cobra_model_from_vmh_recon2, create_cobra_model_from_agora_model
from .kegg.kegg import get_kegg_records, list_kegg_ids, get_kegg_reactions, get_kegg_metabolites, \
    get_kegg_enzymes, get_kegg_amino_acid_seq, get_kegg_dna_seq
from .compare import compare_models, compare_reactions, compare_metabolites, compare_genes




def translate(model, reaction_xref_file_name, metabolite_xref_file_name, from_namespace, to_namespace):
    """ Translate IDs in a model from one namespace to another namespace.

    Parameters
    ----------
    model : cobra.core.Model
        Model to translate
    reaction_xref_file_name : str
        Path to cross reference file with ID mapping for reactions
    metabolite_xref_file_name : str
        Path to cross reference file with ID mapping for metabolites
    from_namespace : str
        Namespace of IDs in input model
    to_namespace : str
        Namespace of IDs in output model

    Returns
    -------
    cobra.core.Model
        New model with translated IDs
    """

    return



import pytest
import cobrababel


class TestVirtualMetabolicHuman:
    def test_create_model(self):
        model = cobrababel.create_cobra_model_from_agora_model('Bacteroides_thetaiotaomicron_VPI_5482')
        assert len(model.metabolites) == 1146
        assert len(model.reactions) == 1362
        assert len(model.genes) == 840
        solution = model.optimize()
        assert solution.f == pytest.approx(0.394520)

    def test_create_recon2(self):
        recon2 = cobrababel.create_cobra_model_from_vmh_recon2()
        assert len(recon2.metabolites) >= 5063
        assert len(recon2.reactions) >= 7440
        assert len(recon2.genes) >= 2140
        solution = recon2.optimize()
        assert solution.f == pytest.approx(3.198056)




from warnings import warn


class KeggReaction:
    
    """ A KEGG reaction record contains the following fields:

        Entry   The KEGG REACTION database is a manually curated collection of biochemical reactions,
                mostly enzymatic reactions but including some spontaneous reactions, which are derived
                from KEGG ENZYME (Enzyme Nomenclature) or KEGG PATHWAY (KEGG metabolic pathway maps).
                Each entry is identified by the unique identifier called the R number ('R' followed
                by five-digit number).
        Name    The name of the reaction, usually the systematic name of the enzyme that catalyzes the
                reaction (can be not specified).
        Definition    The chemical reaction in the form of an equation. The reaction is assumed to be
                reversible and reactants (substrates and products) are separated by '<=>'. Each compound
                in the left or the right side is separated by ' + '. There may be a coefficient before
                the compound name.
        Equation    The C number representation of the reaction equation with links to the COMPOUND
                database entries.
        Remark  The same reaction entries, if any, are given.
        Comment Text information commenting on the reaction.
        RPair   Links to the corresponding KEGG RPAIR database entries, which contain chemical structure
                transformation patterns of reactant pairs (substrate-product pairs).
        Reaction class  Links to the corresponding KEGG RCLASS database entries, which contain reaction
                class information defined by chemical structure transformation patterns of substrate-
                product pairs.
        Enzyme  Links to the corresponding KEGG ENZYME database entries.
        Pathway Links to the KEGG pathway maps, where the corresponding reaction is marked in red.
        Orthology    Links to the corresponding KEGG ORTHOLOGY (KO) database entries.
        Module  Links to the corresponding KEGG MODULE database entries.
        Reference    References for the reaction.
        DBLinks Links to outside databases
    """
    
    def __init__(self, record=None):
        """ Initialize object.

        Parameters
        ----------
        record : list of str, optional
            List of lines with reaction record from flat file database
        """

        self.id = None
        self.name = list()
        self.definition = None
        self.equation = None
        self.remark = None
        self.comment = list()
        self.rpair = list()
        self.rclass = list()
        self.enzyme = list()
        self.pathway = list()
        self.orthology = list()
        self.reference = list()
        self.module = list()
        self.dblinks = list()

        if record is not None:
            self.parse(record)
        return
    
    def as_dict(self):
        """ Return the Reaction object as a dictionary.

        Returns
        -------
        dict
            Dictionary representing Reaction object
        """

        reaction = dict()
        if self.id is not None:
            reaction['id'] = self.id
        if len(self.name) > 0:
            reaction['name'] = self.name
        if self.definition is not None:
            reaction['definition'] = self.definition
        if self.equation is not None:
            reaction['equation'] = self.equation
        if self.remark is not None:
            reaction['remark'] = self.remark
        if len(self.comment) > 0:
            reaction['comment'] = self.comment
        if len(self.rpair) > 0:
            reaction['rpair'] = self.rpair
        if len(self.rclass) > 0:
            reaction['rclass'] = self.rclass
        if len(self.enzyme) > 0:
            reaction['enzyme'] = self.enzyme
        if len(self.pathway) > 0:
            reaction['pathway'] = self.pathway
        if len(self.orthology) > 0:
            reaction['orthology'] = self.orthology
        if len(self.reference) > 0:
            reaction['reference'] = self.reference
        if len(self.module) > 0:
            reaction['module'] = self.module
        if len(self.dblinks) > 0:
            reaction['dblinks'] = self.dblinks

        return reaction
    
    def parse(self, record):
        """ Parse a record from the flat file database to complete this object.

        Parameters
        ----------
        record : list of str
            List of lines with reaction record from flat file database
        """

        for line in record:
            if line[:3] == '///':  # End of record delimiter
                return
            
            # A field in the record has the name in the first 12 characters of the line.
            # If the 12 character prefix is blank, this line is a part of the current field.
            if line[:12] != '            ':
                field_name = line[:12].strip()
            value = line[12:].strip()
            
            # Process each field in the record.
            if field_name == 'ENTRY':
                parts = value.split()
                self.id = parts[0]
            elif field_name == 'NAME':
                # A continuation of the field is delimited by a semicolon at the end of the line.
                self.name.append(value.strip(';'))
            elif field_name == 'DEFINITION':
                self.definition = value
            elif field_name == 'EQUATION':
                self.equation = value
            elif field_name == 'REMARK':
                self.remark = value
            elif field_name == 'COMMENT':
                self.comment.append(value)
            elif field_name == 'RPAIR':
                parts = value.split()
                self.rpair.append(tuple(parts))
            elif field_name == 'RCLASS':
                parts = value.split()
                self.rclass.append(tuple(parts))
            elif field_name == 'ENZYME':
                # Multiple enzymes are all listed on the same line.
                self.enzyme.extend(value.split())
            elif field_name == 'PATHWAY':
                self.pathway.append([value[:7], value[8:].strip()])
            elif field_name == 'ORTHOLOGY':
                self.orthology.append([value[:6], value[7:].strip()])
            elif field_name == 'REFERENCE':
                self.reference.append(value)
            elif field_name == 'MODULE':
                self.module.append([value[:7], value[9:].strip()])
            elif field_name == 'DBLINKS':
                self.dblinks.append(value)
            else:
                warn('Skipping field {0} with value {1}'.format(field_name, value))
        return

    def make_record(self):
        """ Make an reaction record for the flat file database.

        Returns
        -------
        list of str
            List of lines with reaction record
        """

        record = list()
        record.append('ENTRY       {0}                      Reaction'.format(self.id))
        if len(self.name) > 0:
            line = 'NAME        {0}'.format(self.name[0])
            if len(self.name) == 1:
                record.append(line)
            else:
                record.append(line+';')
                for index in range(1, len(self.name)-1):
                    record.append('            {0};'.format(self.name[index]))
                record.append('            {0}'.format(self.name[-1]))
        if self.definition is not None:
            record.append('DEFINITION  {0}'.format(self.definition))
        if self.equation is not None:
            record.append('EQUATION    {0}'.format(self.equation))
        if self.remark is not None:
            record.append('REMARK      {0}'.format(self.remark))
        if len(self.comment) > 0:
            record.append('COMMENT     {0}'.format(self.comment[0]))
            for index in range(1, len(self.comment)):
                record.append('            {0}'.format(self.comment[index]))
        if len(self.rpair) > 0:
            line = 'RPAIR       {0}  '.format(self.rpair[0][0])
            line += ' '.join(self.rpair[0][1:])
            record.append(line)
            for index in range(1, len(self.rpair)):
                line = '            {0}  '.format(self.rpair[index][0])
                line += ' '.join(self.rpair[index][1:])
                record.append(line)
        if len(self.rclass) > 0:
            line = 'RCLASS      {0}  '.format(self.rclass[0][0])
            line += ' '.join(self.rclass[0][1:])
            record.append(line)
            for index in range(1, len(self.rclass)):
                line = '            {0}  '.format(self.rclass[index][0])
                line += ' '.join(self.rclass[index][1:])
                record.append(line)
        if len(self.enzyme) > 0:
            enzymes = '       '.join(self.enzyme)
            record.append('ENZYME      '+enzymes)
        if len(self.pathway) > 0:
            record.append('PATHWAY     {0}  {1}'.format(self.pathway[0][0], self.pathway[0][1]))
            for index in range(1, len(self.pathway)):
                record.append('            {0}  {1}'.format(self.pathway[index][0], self.pathway[index][1]))
        if len(self.orthology) > 0:
            record.append('ORTHOLOGY   {0}  {1}'.format(self.orthology[0][0], self.orthology[0][1]))
            for index in range(1, len(self.orthology)):
                record.append('            {0}  {1}'.format(self.orthology[index][0], self.orthology[index][1]))
        if len(self.reference) > 0:
            record.append('REFERENCE   {0}'.format(self.reference[0]))
            for index in range(1, len(self.reference)):
                record.append('            {0}'.format(self.reference[index]))
        if len(self.module) > 0:
            record.append('MODULE      {0}  {1}'.format(self.module[0][0], self.module[0][1]))
            for index in range(1, len(self.module)):
                record.append('            {0}  {1}'.format(self.module[index][0], self.module[index][1]))
        if len(self.dblinks) > 0:
            record.append('DBLINKS     {0}'.format(self.dblinks[0]))
            for index in range(1, len(self.dblinks)):
                record.append('            {0}'.format(self.dblinks[index]))
        record.append('///')
        return record



from cobra.core import DictList


class RecordNotFound(Exception):
    """ Exception raised when a record is not found. """
    pass


class DatabaseError(Exception):
    """ Exception raised when there is a problem with a database. """
    pass


class KeggDatabase(object):
    """ Base class for managing a KEGG flat file database. """

    def __init__(self, filename):
        """ Initialize object.

        Parameters
        ----------
        filename : str
            Path to database file
        """

        self.filename = filename
        self.records = DictList()
        return
    
    def get_record(self, handle):
        """ Get a record from a database file.
        
        Parameters
        ----------
        handle : file handle
            File handle of database file

        Returns
        -------
        list of str
             List of lines in record
        """
        
        record = list()
        for line in handle:
            record.append(line.strip('\n'))
            if line[:3] == '///':
                yield record
                record = list()
                continue

    def store(self, filename=None):
        """ Save the database to a flat file.

        Parameters
        ----------
        filename : str, optional
            Path to database file
\       """

        if filename is None:
            filename = self.filename

        # Convert all of the record objects to flat file database records and write to the file.
        self.records.sort()
        with open(filename, 'w') as handle:
            for index in range(len(self.records)):
                for line in self.records[index].make_record():
                    handle.write(line + '\n')
        return

    def update(self, new_object):
        """ Update a record in the database (add new or replace existing record).

        Parameters
        ----------
        new_object : object
            Record object to add or replace
        """

        # Replace the current object if it already exists in the database.
        if self.records.has_id(new_object.id):
            self.records._replace_on_id(new_object)
            return

        # Add the new object to the database.
        self.records += [new_object]
        return

    def size(self):
        """ Get the number of records in the database.

        Returns
        -------
        int
            Number of records in database
        """

        return len(self.records)

    def has_id(self, id):
        """ Check if an ID exists in the database.

        Parameters
        ----------
        id : str
            ID to check

        Returns
        -------
        bool
            True when ID exists, otherwise False
        """

        return self.records.has_id(id)

    def get_by_id(self, id):
        """ Get an record with the specified ID.

        Parameters
        ----------
        id : str
            ID of record to return

        Returns
        -------
        object
            Object with specified ID
        """

        return self.records.get_by_id(id)



import requests
import io
from warnings import warn
from time import sleep
import logging

from cobra.core import Model, Metabolite, Reaction, DictList
from cobra.io import load_json_model

# Base URL for BiGG website
bigg_url = 'http://bigg.ucsd.edu/api/v2/'

# Pause after this many requests to BiGG data API
PAUSE_COUNT = 500

# Logger for this module
LOGGER = logging.getLogger(__name__)


def create_bigg_universal_model(validate=False):
    """ Create an universal model from BiGG universal reactions and metabolites.

    Parameters
    ----------
    validate : bool, optional
        When True, perform validity checks on universal COBRA model

    Returns
    -------
    cobra.core.Model
        COBRA model object with universal reactions and metabolites
    """

    # Get the current version number for the BiGG database.
    response = requests.get(bigg_url + 'database_version')
    if response.status_code != requests.codes.OK:
        response.raise_for_status()
    version = response.json()

    # Create an empty model.
    universal = Model('bigg_universal', name='BiGG universal model {0}'.format(version['bigg_models_version']))
    universal.notes['last_updated'] = version['last_updated']

    # Get the list of universal metabolites.
    LOGGER.info('Started download of universal metabolite list')
    response = requests.get(bigg_url + 'universal/metabolites')
    if response.status_code != requests.codes.OK:
        response.raise_for_status()
    metabolite_list = response.json()['results']
    LOGGER.info('Finished download of universal metabolite list')

    # Get the details on each universal metabolite. Note there is no bulk download
    # of universal metabolites so there is separate request to the BiGG server
    # for each metabolite which makes this very slow.
    LOGGER.info('Started download of %d metabolites', len(metabolite_list))
    metabolites = list()
    for index in range(len(metabolite_list)):
        if index % PAUSE_COUNT == 0:
            LOGGER.info('Paused at metabolite number %s', index)
            sleep(1)  # Be nice to BiGG data API and take a breath
        bigg_metabolite = get_bigg_metabolite(metabolite_list[index]['bigg_id'])

        # Add a metabolite for each unique compartment.
        compartment_list = set(x['bigg_id'] for x in bigg_metabolite['compartments_in_models'])
        for compartment in compartment_list:
            bigg_metabolite['compartment_bigg_id'] = compartment
            metabolites.append(bigg_metabolite.copy())
    LOGGER.info('Finished download of metabolites')

    # Add the metabolites to the universal model.
    LOGGER.info('Started adding %d Metabolite objects to universal model', len(metabolites))
    add_bigg_metabolites(metabolites, universal)
    LOGGER.info('Finished adding Metabolite objects to universal model')

    # Get the list of universal reactions.
    LOGGER.info('Started download of universal reaction list')
    response = requests.get(bigg_url + 'universal/reactions')
    if response.status_code != requests.codes.OK:
        response.raise_for_status()
    reaction_list = response.json()['results']
    LOGGER.info('Finished download of universal reaction list')

    # Get the details on each universal reaction. Remember there is no bulk download
    # of universal reactions which makes this very slow.
    LOGGER.info('Started creating Reaction objects for %d reactions', len(reaction_list))
    reactions = list()
    for index in range(len(reaction_list)):
        if index % PAUSE_COUNT == 0:
            LOGGER.info('Paused at reaction number %s', index)
            sleep(1)  # Be nice to BiGG data API and take a breath
        reactions.append(get_bigg_reaction(reaction_list[index]['bigg_id']))
    LOGGER.info('Finished creating %d reaction objects', len(reactions))

    # Add the reactions to the universal model.
    LOGGER.info('Started adding %d Reaction objects to universal model', len(reactions))
    add_bigg_reactions(reactions, universal)
    LOGGER.info('Finished adding Reaction objects to universal model')

    # If requested, validate the COBRA model.
    if validate:
        warn('Coming soon')

    return universal


def create_cobra_model_from_bigg_model(bigg_id, validate=False):
    """ Create a COBRA model from a BiGG model.

    Parameters
    ----------
    bigg_id: str
        ID of BiGG model
    validate : bool, optional
        When True, perform validity checks on COBRA model

    Returns
    -------
    cobra.core.Model
        COBRA model created from JSON representation of BiGG model
    """

    # Download the JSON representation and details of the model from BiGG.
    LOGGER.info('Started download of %s model', bigg_id)
    response = requests.get('{0}models/{1}'.format(bigg_url, bigg_id))
    if response.status_code != requests.codes.OK:
        response.raise_for_status()
    details = response.json()

    response = requests.get('{0}models/{1}/download'.format(bigg_url, bigg_id))
    if response.status_code != requests.codes.OK:
        response.raise_for_status()
    LOGGER.info('Finished download of %s model', bigg_id)

    # Convert to a cobra.Model object.
    with io.StringIO(response.text) as f:
        model = load_json_model(f)

    # Add some details to the Model object.
    model.name = details['organism']
    model.notes['genome_name'] = details['genome_name']
    model.notes['reference_type'] = details['reference_type']
    model.notes['reference_id'] = details['reference_id']

    # Confirm a few basics.
    if len(model.reactions) != details['reaction_count']:
        warn('{0} reactions in model does not equal {1} in model details'
             .format(len(model.reactions), details['reaction_count']))
    if len(model.metabolites) != details['metabolite_count']:
        warn('{0} metabolites in model does not equal {1} in model details'
             .format(len(model.metabolites), details['metabolite_count']))
    if len(model.genes) != details['gene_count']:
        warn('{0} genes in model does not equal {1} in model details'
             .format(len(model.genes), details['gene_count']))

    # If requested, validate the COBRA model.
    if validate:
        warn('Coming soon')

    return model


def get_bigg_model_list():
    """ Get the list of models available from BiGG website.

    Each entry in the list is a dictionary with keys that include 'bigg_id',
    'organism', 'metabolite_count', 'reaction_count', and 'gene_count'.

    Returns
    -------
    list of dict
        List of models available from BiGG website
    """

    LOGGER.info('Started download of model list')
    response = requests.get(bigg_url + 'models')
    if response.status_code != requests.codes.OK:
        response.raise_for_status()
    LOGGER.info('Finished download of model list')
    return response.json()['results']


def get_bigg_metabolite(bigg_id, model_bigg_id='universal'):
    """ Get a metabolite from the BiGG database.

    Parameters
    ----------
    bigg_id: str
        ID of BiGG metabolite
    model_bigg_id : str, optional
        ID of model containing metabolite

    Returns
    -------
    dict
        Dictionary with metabolite data
    """

    if model_bigg_id == 'universal':
        url = '{0}universal/metabolites/{1}'.format(bigg_url, bigg_id)
    else:
        url = '{0}models/{1}/metabolites/{2}'.format(bigg_url, model_bigg_id, bigg_id)
    LOGGER.debug('Started download of %s metabolite from %s model', bigg_id, model_bigg_id)
    response = requests.get(url)
    if response.status_code != requests.codes.OK:
        response.raise_for_status()
    LOGGER.debug('Finished download of %s metabolite from %s model', bigg_id, model_bigg_id)
    return response.json()


def add_bigg_metabolites(bigg_list, model):
    """ Create a COBRA metabolite from a BiGG metabolite.

    Parameters
    ----------
    bigg_list: list of dict
        List of dictionaries with BiGG metabolite data
    model: cobra.core.Model
        Model to add metabolites to
    """

    # Create a Metabolite object for each BiGG metabolite.
    metabolites = DictList()
    for bigg_metabolite in bigg_list:
        # Available data is different for a metabolite from an organism model versus
        # a metabolite from the universal model.
        if 'compartment_bigg_id' in bigg_metabolite:
            compartment = bigg_metabolite['compartment_bigg_id']
        elif 'compartments_in_models' in bigg_metabolite:
            compartment = bigg_metabolite['compartments_in_models'][0]['bigg_id']
        else:
            raise ValueError('BiGG metabolite {0} does not have a compartment'.format(bigg_metabolite['bigg_id']))
        metabolite = Metabolite(id='{0}_{1}'.format(bigg_metabolite['bigg_id'], compartment),
                                name=bigg_metabolite['name'],
                                compartment=compartment)
        try:
            metabolite.formula = bigg_metabolite['formula']
        except KeyError:
            try:
                if len(bigg_metabolite['formulae']) > 0:
                    metabolite.formula = bigg_metabolite['formulae'][0]
            except KeyError:
                pass
        try:
            metabolite.charge = bigg_metabolite['charge']
        except KeyError:
            try:
                if len(bigg_metabolite['charges']) > 0:
                    metabolite.charge = bigg_metabolite['charges'][0]
            except KeyError:
                pass
        if len(bigg_metabolite['database_links']) > 0:
            metabolite.notes['aliases'] = bigg_metabolite['database_links']
        metabolites.append(metabolite)

        if compartment not in model.compartments:
            try:
                model.compartments[compartment] = bigg_metabolite['compartment_name']
            except KeyError:
                model.compartments[compartment] = 'unknown'

    # Add all of the metabolites to the model.
    model.add_metabolites(metabolites)
    return


def get_bigg_reaction(bigg_id, model_bigg_id='universal'):
    """ Get a reaction from the BiGG database.

    Parameters
    ----------
    bigg_id: str
        ID of BiGG reaction
    model_bigg_id : str, optional
        ID of model containing reaction

    Returns
    -------
    dict
        Dictionary with reaction data
    """

    if model_bigg_id == 'universal':
        url = '{0}universal/reactions/{1}'.format(bigg_url, bigg_id)
    else:
        url = '{0}models/{1}/reactions/{2}'.format(bigg_url, model_bigg_id, bigg_id)
    LOGGER.debug('Started download of reaction %s from model %s', bigg_id, model_bigg_id)
    response = requests.get(url)
    if response.status_code != requests.codes.OK:
        response.raise_for_status()
    LOGGER.debug('Finished download of reaction %s from model %s', bigg_id, model_bigg_id)
    return response.json()


def add_bigg_reactions(bigg_list, model):
    """ Create a COBRA reaction from a BiGG reaction.

    Parameters
    ----------
    bigg_list: list of dict
        List of dictionaries with BiGG reaction data
    model: cobra.core.Model
        Model to add reactions to
    """

    # Create a Reaction object for each BiGG reaction.
    reactions = DictList()
    for bigg_reaction in bigg_list:
        reaction = Reaction(id=bigg_reaction['bigg_id'], name=bigg_reaction['name'])
        reaction.notes['aliases'] = bigg_reaction['database_links']
        metabolites = dict()
        for met in bigg_reaction['metabolites']:
            metabolite = model.metabolites.get_by_id('{0}_{1}'.format(met['bigg_id'], met['compartment_bigg_id']))
            metabolites[metabolite] = met['stoichiometry']
        reaction.add_metabolites(metabolites)
        try:
            reaction.bounds = (bigg_reaction['results'][0]['lower_bound'], bigg_reaction['results'][0]['upper_bound'])
        except KeyError:
            if '&#8652' in bigg_reaction['reaction_string']:
                reaction.bounds = (-1000.0, 1000.0)
            else:
                warn('Unknown direction symbol in reaction string {0}'.format(bigg_reaction['reaction_string']))
        reactions.append(reaction)

    # Add all of the reactions to the model.
    model.add_reactions(reactions)
    return



from __future__ import absolute_import

import json
import jsonschema
import re
from warnings import warn
from math import log10


def save_visual_json_model(model, file_name, pathways=None, pretty=False, threshold=1E-8):
    """ Save a model to a JSON file for MetabolicNetworks visualization.

    Parameters
    ----------
        model
            Model to analyze (must be optimized)
        file_name : str
            Path to file for saving MetabolicNetworks JSON
        pathways : dict, optional
            Dictionary that identifies what pathways a reaction participates in
        pretty : bool, optional
            When True, pretty print the JSON
        threshold : float, optional
            Tolerance for determining if a flux is zero
    """

    # First check for ModelSEED format compartment suffix. A ModelSEED model uses
    # compartment suffixes in the format "_c0" or "_c".
    compartment_suffix_re = re.compile(r'_([a-z])\d*$')
    match = re.search(compartment_suffix_re, model.reactions[0].id)
    if match is None:
        # Second check for BiGG format compartment suffix. A BiGG model uses compartment
        # suffixes in the format "[c]".
        compartment_suffix_re = re.compile(r'\[([a-z])\]$')
        match = re.search(compartment_suffix_re, model.reactions[0])
        if match is None:
            raise ValueError('Unknown compartment suffix format')
            # @todo Could get_kegg_records around this if compartments are set correctly?

    # Set the ID of the cytosol compartment in the model.
    # @todo What if compartments are not set?
    cytosol = None
    for compartment_id in model.compartments:
        if model.compartments[compartment_id].lower() == 'cytosol':
            cytosol = compartment_id

    # Set compartment on all metabolites if it is not already set.
    for index in range(len(model.metabolites)):
        metabolite = model.metabolites[index]
        if metabolite.compartment is None:
            # @todo What if the metabolite does not have a compartment suffix?
            metabolite.compartment = re.search(compartment_suffix_re, metabolite.id).group(1)

    # Look for boundary reactions to find metabolites that are being consumed and produced.
    consumed_metabolites = set()
    produced_metabolites = set()
    boundary_reactions = model.reactions.query(lambda x: x, 'boundary')
    for reaction in boundary_reactions:
        if len(reaction.reactants) > 0:  # @todo Why do I care about this?
            metabolite = reaction.reactants[0]
        else:
            print('Boundary reaction {0}: {1} has product metabolite and was skipped'.
                  format(reaction.id, reaction.reaction))
            print('  Flux is {0}'.format(reaction.x))
            continue

        # A positive flux means the metabolite is consumed after adjusting for metabolite stoichiometry.
        # Reminder: reaction.x is negative for consumed metabolites.
        flux = reaction.x * reaction.metabolites[metabolite]
        if flux > threshold:
            consumed_metabolites.add(metabolite.id)
        elif flux < threshold:
            produced_metabolites.add(metabolite.id)

    # Create a metabolite structure for each metabolite in the model.
    metabolites = dict()
    for metabolite in model.metabolites:
        metabolite_data = dict()
        metabolite_data['id'] = metabolite.id
        if metabolite.name is not None:
            metabolite_data['name'] = metabolite.name
        else:
            metabolite_data['name'] = metabolite.id
        metabolite_data['name'] = re.sub(compartment_suffix_re, '', metabolite_data['name'])
        metabolite_data['formula'] = metabolite.formula
        metabolite_data['compound'] = re.sub(compartment_suffix_re, '', metabolite.id)
        if metabolite.id in consumed_metabolites:
            metabolite_data['media'] = 1
        else:
            metabolite_data['media'] = 0
        if metabolite.id in produced_metabolites:
            metabolite_data['is_output'] = 1
        else:
            metabolite_data['is_output'] = 0
        metabolite_data['compartment'] = metabolite.compartment + '0'
        metabolites[metabolite.id] = metabolite_data

    # Find maximum upper bound.
    max_upper_bound = max(model.reactions.list_attr("upper_bound"))

    # Create the representation of the model for visualization.
    visual_model = dict()
    name_words = model.name.split(' ')  # Used to generate short name
    visual_model['name'] = {'short': name_words[0][0]+name_words[1][0:2], 'long': model.name}

    # Identify the cofactor metabolites based on the source of the model. The values come
    # from the database used to build the model (e.g. ModelSEED or BiGG).
    if len(model.metabolites.query('cpd00001_c', 'id')) > 0:
        visual_model['cofactors'] = [
            'H2O', 'H+', 'Phosphate', 'ATP', 'ADP',
            'GTP', 'GDP', 'CO2', 'NADH', 'NADHP',
            'Oxidizedferredoxin', 'Reducedferredoxin', 'FAD', 'FADH2'
        ]
    elif len(model.metabolites.query('h2o', 'id')) > 0:
        visual_model['cofactors'] = [
            'h2o', 'h', 'pi', 'atp', 'adp',
            'gtp', 'gdp', 'co2', 'nadh', 'nadph',
            'fad', 'fadh2'
        ]

    # The list of nodes describes all of the reactions in the model.
    visual_model['nodes'] = list()
    use_names = True  # @todo Need to detect if names are available in metabolites
    num_active_reactions = 0
    for reaction in model.reactions:
        # Boundary reactions were handled above so we can skip them here.
        if reaction.boundary:
            continue

        # Make a reaction node.
        reaction_data = dict()
        reaction_data['node_type'] = 'reaction'
        reaction_data['id'] = reaction.id
        reaction_data['reaction'] = re.sub(compartment_suffix_re, '', reaction.id)
        if reaction.name != '':
            reaction_data['name'] = reaction.name
        else:
            reaction_data['name'] = 'unknown'
        reaction_data['definition'] = reaction.build_reaction_string(use_metabolite_names=use_names)
        reaction_data['flux'] = reaction.x
        if 'gapfill_data' in reaction.notes:
            reaction_data['gapfilled'] = 'Yes'
        else:
            reaction_data['gapfilled'] = 'No'
        if cytosol in reaction.get_compartments():
            # Any metabolite in cytosol makes the reaction in cytosol.
            reaction_data['compartment'] = 'c0'
        else:
            # Microbes are simple -- anything else is extracellular.
            reaction_data['compartment'] = 'e0'

        # Reaction direction is set based on lower and upper bounds.
        # @todo Does this need to be adjusted for fva?
        if reaction.lower_bound < 0.0 and reaction.upper_bound > 0.0:
            direction = '='
        elif reaction.lower_bound >= 0.0 and reaction.upper_bound > 0.0:
            direction = '>'
        elif reaction.lower_bound < 0.0 and reaction.upper_bound <= 0.0:
            direction = '<'
        else:
            warn('Could not set direction from lower bound {0} and upper bound {1} for reaction {2}'.
                 format(reaction.lower_bound, reaction.upper_bound, reaction.id))
            continue

        # Features is a list of gene IDs.
        # @todo Remove features because not used in visualization?
        reaction_data['features'] = [gene.id for gene in reaction.genes]

        # Add a metabolite data structure for each metabolite in the reaction.
        reactants = list()
        for metabolite in reaction.reactants:
            metabolite_data = metabolites[metabolite.id]
            metabolite_data['stoich'] = reaction.metabolites[metabolite]
            reactants.append(metabolite_data)
        products = list()
        for metabolite in reaction.products:
            metabolite_data = metabolites[metabolite.id]
            metabolite_data['stoich'] = reaction.metabolites[metabolite]
            products.append(metabolite_data)

        # Sometimes the solver gets a solution close to zero. Round the flux to zero when it is very close.
        if reaction_data['flux'] != 0.0 and abs(reaction_data['flux']) < threshold:
            reaction_data['flux'] = 0.0
        if reaction_data['flux'] != 0.0:
            num_active_reactions += 1

        # For unidirectonal reactions, make sure the flux value is consistent.
        if direction == '>' and reaction_data['flux'] < 0.0:
            warn('Negative flux on forward reaction {0}'.format(reaction_data))
        if direction == '<' and reaction_data['flux'] > 0.0:
            warn('Positive flux on reverse reaction {0}'.format(reaction_data))

        # For reverse reactions, flip the reactants and products and the
        # symbol in the definition.
        if direction == '<':
            reaction_data['reactants'] = products
            reaction_data['products'] = reactants
            parts = reaction_data['definition'].split(' <-- ')
            reaction_data['definition'] = '{0} --> {1}'.format(parts[1], parts[0])

        # For bidirectional reactions, flip the reactants and products when the
        # flux value is negative.
        elif direction == '=':
            if reaction_data['flux'] != 0.0:
                if reaction_data['flux'] < 0.0:
                    reaction_data['reactants'] = products
                    reaction_data['products'] = reactants
                    parts = reaction_data['definition'].split(' <=> ')
                    reaction_data['definition'] = '{0} --> {1}'.format(parts[1], parts[0])
                else:
                    reaction_data['reactants'] = reactants
                    reaction_data['products'] = products
                    reaction_data['definition'] = reaction_data['definition'].replace('<=>', '-->')
            else:
                reaction_data['reactants'] = reactants
                reaction_data['products'] = products

        # For forward reactions there are no changes.
        else:
            reaction_data['reactants'] = reactants
            reaction_data['products'] = products

        # Now the flux value can always be a positive value.
        reaction_data['flux'] = abs(reaction_data['flux'])

        # Scale a non-zero flux value to fit in the range 0-100. There is a limitation in
        # MetabolicNetworks that requires the flux range to be fixed.
        # @todo Pick a different transform based on the max upper bound value?
        if reaction_data['flux'] != 0.0 and max_upper_bound > 100.0:
            # COBRA models have a maximum upper bound for reaction flux of 1000.
            # Flux range is transformed from 0-1000 to 0-99.015.
            reaction_data['flux'] = log10(reaction_data['flux']+1.0) * 33.00

        # Add the pathways that the reaction participates in.
        if pathways is not None:
            try:
                reaction_data['pathway'] = pathways['reactions'][reaction.id]
            except KeyError:
                reaction_data['pathway'] = list()

        # Add the reaction node.
        visual_model['nodes'].append(reaction_data)

    # Add the pathway names if needed.
    if pathways is not None:
        visual_model['pathways'] = pathways['pathways']

    # Create the json file with the visual model.
    if pretty:
        dump_opts = {"indent": 4, "separators": (",", ": "), "sort_keys": True}
    else:
        dump_opts = {}
    jsonschema.validate(visual_model, visual_model_json_schema)  # @todo Move to unit test
    json.dump(visual_model, open(file_name, "w"), **dump_opts)

    return


visual_model_json_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "title": "Visual COBRA",
    "description": "JSON representation of COBRA model for visualization by MetabolicNetworks",
    "type": "object",
    "properties": {
        "name": {
            "type": "object",
            "properties": {
                "short": {"type": "string"},
                "long": {"type": "string"},
            }
        },
        "cofactors": {
            "type": "array",
            "items": {"type": "string"},
        },
        "nodes": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "node_type": {"type": "string"},
                    "id": {"type": "string"},
                    "reaction": {"type": "string"},  # Is this necessary?
                    "name": {"type": "string"},
                    "definition": {"type": "string"},
                    "flux": {"type": "number"},
                    "gapfilled": {"type": "string", "pattern": "Yes|No"},
                    "compartment": {"type": "string", "pattern": "[a-z]0"},
                    "features": {  # Is this necessary?
                        "type": "array",
                        "items": {"type": "string"},
                    },
                    "pathways": {
                        "type": "array",
                        "items": {"type": "string"},
                    },
                    "reactants": {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "properties": {
                                "id": {"type": "string"},
                                "name": {"type": "string"},
                                "compound": {"type": "string"},  # Is this necessary?
                                "media": {"type": "integer"},
                                "compartment": {"type": "string", "pattern": "[a-z]0"},
                                "is_output": {"type": "integer"},
                                "formula": {"type": "string"},
                                "stoich": {"type": "number"},  # Is this necessary?
                            },
                        },
                        "required": ["id", "name", "compound", "media", "compartment"],
                    },
                    "products": {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "properties": {
                                "id": {"type": "string"},
                                "name": {"type": "string"},
                                "compound": {"type": "string"},  # Is this necessary?
                                "media": {"type": "integer"},
                                "compartment": {"type": "string", "pattern": "[a-z]0"},
                                "is_output": {"type": "integer"},
                                "formula": {"type": "string"},
                                "stoich": {"type": "number"},  # Is this necessary?
                            },
                        },
                        "required": ["id", "name", "compound", "media", "compartment"],
                    },
                },
                "required": ["id", "name", "reaction", "compartment", "definition", "flux", "reactants", "products"],
            },
        },
        "pathways": {
            "type": "object",
            "properties": {
                "id": {"type": "string"},
                "name": {"type": "string"},
            },
        },
        "version": {
            "type": "integer",
            "default": 1,
        },

    },
    "required": ["name", "nodes"],
    "additionalProperties": False,
}



def format_long_string(string, max_length):
    """ Format a string so it fits in column of a specific width.

    Parameters
    ----------
    string : str
        String to format
    max_length : int
        Maximum length of returned string

    Returns
    -------
    str
        Formatted string
    """

    if len(string) > max_length:
        string = string[:max_length - 3]
        string += '...'
    return string



class KeggEnzyme(object):
    
    """ A KEGG enzyme record contains the following fields:

        Entry    KEGG ENZYME contains the information about Enzyme Nomenclature obtained from the
                 ExplorEnz database. Additional information is included both computationally and
                 manually. Manually added information includes the KEGG reaction data with parent-child
                 (general to more specific) relationship and the source organism and protein sequence
                 information, whenever available, in each reference. Links to various data in KEGG and
                 other databases are computationally generated. The accession number of an ENZYME entry
                 is the EC (Enzyme Commission) number given by the Nomenclature Committee of the IUBMB
                 and the IUPAC. It is preceded by EC here, simply because of a historical reason.
        Name    The accepted name and alternative names of the enzyme.
        Class   The class, subclass, and sub-subclass of the enzyme, which correspond to the first
                three figures of the EC number.
        Sysname    The systematic name of the enzyme given by the Nomenclature Committee. It represents
                the nature of the chemical reaction.
        Reaction(KEGG)    This field contains all the reactions (R numbers) with this EC number in KEGG.
                Additional reactions are either examples of the general reaction formula by IUBMB,
                designated by ">" for the parent-child relationship, or unrelated to the IUBMB reaction
                formula designated by "(other)".
        Substrate  The chemical compounds that appear on the left side of the reaction equation.
        Product    The chemical compounds that appear on the right side of the reaction equation.
        Comment    Text information commenting on the enzyme.
        Pathway    Links to the KEGG pathway maps, where the corresponding enzyme is marked in red.
        Orthology  Link to the K number entry in the KEGG Orthology (KO) database, which corresponds to
                the ortholog group for the enzyme.
        Genes   Links to the GENES database entries with the assignment (through the KO system) of the
                corresponding EC number.
        Reaction(IUBMB)    The chemical reaction catalyzed by the enzyme is shown in the form of an
                equation or in a text description as defined by the IUBMB/IUPAC Nomenclature Committee.
                The corresponding R number in the KEGG REACTION database is also shown.
        History    Update history given by the IUBMB Enzyme Nomenclature Committee
        Reference  References describing the enzyme.
    """
    
    def __init__(self, record=None):
        """ Initialize object.

        Parameters
        ----------
        record : list of str, optional
            List of lines with enzyme record from flat file database
        """

        self.id = None
        self.name = list()
        self.obsolete = 0
        self.classname = list()
        self.sysname = None
        self.reaction = list()
        self.rxnid = list()
        self.allreactions = list()
        self.substrate = list()
        self.product = list()
        self.comment = list()
        self.pathway = list()
        self.orthology = list()
        self.genes = dict()

        if record is not None:
            self.parse(record)
        return
    
    def as_dict(self):
        """ Return the Enzyme object as a dictionary.
        
        Returns
        -------
        dict
            Dictionary representing Enzyme object
        """

        enzyme = dict()
        if self.id is not None:
            enzyme['id'] = self.id
        enzyme['obsolete'] = self.obsolete
        if len(self.name) > 0:
            enzyme['name'] = self.name
        if len(self.classname) > 0:
            enzyme['classname'] = self.classname
        if self.sysname is not None:
            enzyme['sysname'] = self.sysname
        if len(self.reaction) > 0:
            enzyme['reaction'] = self.reaction
        if len(self.allreactions) > 0:
            enzyme['allreactions'] = self.allreactions
        if len(self.substrate) > 0:
            enzyme['substrate'] = self.substrate
        if len(self.product) > 0:
            enzyme['product'] = self.product
        if len(self.comment) > 0:
            enzyme['comment'] = self.comment
        if len(self.pathway) > 0:
            enzyme['pathway'] = self.pathway
        if len(self.orthology) > 0:
            enzyme['orthology'] = self.orthology
        if len(self.genes) > 0:
            enzyme['genes'] = self.genes
        
        return enzyme
    
    def parse(self, record):
        """ Parse a record from the flat file database to complete this object.

        Parameters
        ----------
        record : list of str
            List of lines with enzyme record from flat file database
        """
        
        for line in record:
            if line[:3] == '///':  # End of record delimiter
                return
            
            # A field in the record has the name in the first 12 characters of the line.
            # If the 12 character prefix is blank, this line is a part of the current field.
            if line[:12] != '            ':
                field_name = line[:12].strip()
            value = line[12:].strip()
            
            # Process each field in the record.
            if field_name == 'ENTRY':
                # Entry is in the format EC n.n.n.n
                parts = value.split()
                self.id = parts[1]
                if parts[2] == 'Obsolete':
                    self.obsolete = 1
            elif field_name == 'NAME':
                # A continuation of the field is delimited by a semicolon at the end of the line.
                self.name.append(value.strip(';'))
            elif field_name == 'CLASS':
                self.classname.append(value.strip(';'))
            elif field_name == 'SYSNAME':
                self.sysname = value
            elif field_name == 'REACTION':
                # Extract the reaction numbers for easy cross referencing.
                val = value.strip(';')
                start_pos = val.find('[RN:')
                if start_pos >= 0:
                    id_list = val[start_pos+4:-1]
                    self.rxnid.extend(id_list.split())
                self.reaction.append(val)
            elif field_name == 'ALL_REAC':
                self.allreactions.append(value.strip(';'))
            elif field_name == 'SUBSTRATE':
                self.substrate.append(value.strip(';'))
            elif field_name == 'PRODUCT':
                self.product.append(value.strip(';'))
            elif field_name == 'COMMENT':
                self.comment.append(value)
            elif field_name == 'PATHWAY':
                self.pathway.append([value[:7], value[8:].strip()])
            elif field_name == 'ORTHOLOGY':
                self.orthology.append([value[:6], value[7:].strip()])
            elif field_name == 'GENES':
                # Field has organism code, a colon, and a space separated list of genes
                pos = value.find(':')
                organism = value[:pos].lower()
                self.genes[organism] = value[pos+2:].split()
            
        return   

    def make_record(self):
        """ Make an enzyme record for the flat file database.

        Returns
        -------
        list of str
            List of lines with enzyme record
        """

        record = list()
        entry = 'ENTRY       EC {0}'.format(self.id)
        if self.obsolete:
            pad = 30 - len(entry)
            for index in range(pad):
                entry += ' '
            entry += 'Obsolete  Enzyme'
        else:
            pad = 40 - len(entry)
            for index in range(pad):
                entry += ' '
            entry += 'Enzyme'
        record.append(entry)
        if len(self.name) > 0:
            line = 'NAME        {0}'.format(self.name[0])
            if len(self.name) == 1:
                record.append(line)
            else:
                record.append(line+';')
                for index in range(1, len(self.name)-1):
                    record.append('            {0};'.format(self.name[index]))
                record.append('            {0}'.format(self.name[-1]))
        if len(self.classname) > 0:
            line = 'CLASS       {0}'.format(self.classname[0])
            if len(self.classname) == 1:
                record.append(line)
            else:
                record.append(line+';')
                for index in range(1, len(self.classname)-1):
                    record.append('            {0};'.format(self.classname[index]))
                record.append('            {0}'.format(self.classname[-1]))
        if self.sysname is not None:
            record.append('SYSNAME     {0}'.format(self.sysname))
        if len(self.reaction) > 0:
            line = 'REACTION    {0}'.format(self.reaction[0])
            if len(self.reaction) == 1:
                record.append(line)
            else:
                record.append(line+';')
                for index in range(1, len(self.reaction)-1):
                    record.append('            {0};'.format(self.reaction[index]))
                record.append('            {0}'.format(self.reaction[-1]))
        if len(self.allreactions) > 0:
            line = 'ALL_REAC    {0}'.format(self.allreactions[0])
            if len(self.allreactions) == 1:
                record.append(line)
            else:
                record.append(line+';')
                for index in range(1, len(self.allreactions)-1):
                    record.append('            {0};'.format(self.allreactions[index]))
                record.append('            {0}'.format(self.allreactions[-1]))
        if len(self.substrate) > 0:
            line = 'SUBSTRATE   {0}'.format(self.substrate[0])
            if len(self.substrate) == 1:
                record.append(line)
            else:
                record.append(line+';')
                for index in range(1, len(self.substrate)-1):
                    record.append('            {0};'.format(self.substrate[index]))
                record.append('            {0}'.format(self.substrate[-1]))
        if len(self.product) > 0:
            line = 'PRODUCT     {0}'.format(self.product[0])
            if len(self.product) == 1:
                record.append(line)
            else:
                record.append(line+';')
                for index in range(1, len(self.product)-1):
                    record.append('            {0};'.format(self.product[index]))
                record.append('            {0}'.format(self.product[-1]))
        if len(self.comment) > 0:
            record.append('COMMENT     {0}'.format(self.comment[0]))
            for index in range(1, len(self.comment)):
                record.append('            {0}'.format(self.comment[index]))
        if len(self.pathway) > 0:
            record.append('PATHWAY     {0}  {1}'.format(self.pathway[0][0], self.pathway[0][1]))
            for index in range(1, len(self.pathway)):
                record.append('            {0}  {1}'.format(self.pathway[index][0], self.pathway[index][1]))
        if len(self.orthology) > 0:
            record.append('ORTHOLOGY   {0}  {1}'.format(self.orthology[0][0], self.orthology[0][1]))
            for index in range(1, len(self.orthology)):
                record.append('            {0}  {1}'.format(self.orthology[index][0], self.orthology[index][1]))
        if len(self.genes) > 0:
            first = True
            for key in sorted(self.genes):
                if first:
                    prefix = 'GENES       '
                    first = False
                else:
                    prefix = '            '
                record.append('{0}{1}: {2}'.format(prefix, key, ' '.join(self.genes[key])))
        record.append('///')
        return record



import requests
from warnings import warn
import re
import logging

from cobra import Model, Metabolite, Reaction, DictList

# Base URL for MetaNetX website
metanetx_url = 'http://www.metanetx.org/cgi-bin/mnxget/mnxref/'

# Supported MetaNetX version (corresponds to MNXref 3.0)
metanetx_version = 'MNXref Version 2017/05/04'

# Regular expression for metabolites in reaction equation
metabolite_re = re.compile(r'(\d*\.\d+|\d+) (MNXM\d+|BIOMASS)@(MNXD[\dX]|BOUNDARY)')

# Logger for this module
LOGGER = logging.getLogger(__name__)


def create_metanetx_universal_model(validate=False, verbose=False):
    """ Create an universal model from MetaNetX universal reactions and metabolites.

    The MetaNetX metabolite list is very large and includes metabolites that are
    not used in any reaction. The returned model only includes metabolites that
    are actually used in a reaction.

    Parameters
    ----------
    validate : bool, optional
        When True, perform validity checks on universal COBRA model
    verbose : bool, optional
        When True, show warning messages

    Returns
    -------
    cobra.Model
        COBRA model object with universal reactions and metabolites
    """

    # Create an empty model.
    universal = Model('metanetx_universal', name='MetaNetX universal model')

    # Download the metabolites file.
    metabolite_list = _download_metanetx_file('chem_prop.tsv')

    # Map field names to column numbers (MetaNetX may add fields in future).
    field_names = {
        'MNX_ID': 0,
        'Description': 1,
        'Formula': 2,
        'Charge': 3,
        'Mass': 4,
        'InChI': 5,
        'SMILES': 6,
        'Source': 7,
        'InChIKey': 8
    }

    # Accumulate all available Metabolite objects separately. Later when creating
    # reactions, metabolites are put in a compartment.
    all_metabolites = DictList()

    # Create Metabolite objects for all of the metabolites from the downloaded file.
    # Add all of the universal metabolites from the list.
    LOGGER.info('Started creating Metabolite objects from %d lines in file', len(metabolite_list))
    for index in range(len(metabolite_list)):
        if len(metabolite_list[index]) == 0 or metabolite_list[index][0] == '#':
            continue  # Skip empty lines and comment lines
        fields = metabolite_list[index].split('\t')
        if len(fields) < len(field_names):
            if verbose:
                warn('Skipped metabolite on line {0} with missing fields: {1}'.format(index, metabolite_list[index]))
            continue

        # Create cobra.core.Metabolite from MetaNetX metabolite.
        metabolite = Metabolite(id=fields[field_names['MNX_ID']],
                                name=fields[field_names['Description']],
                                formula=fields[field_names['Formula']])
        charge = fields[field_names['Charge']]
        metabolite.charge = int(charge) if len(charge) > 0 and charge != 'NA' else None
        mass = fields[field_names['Mass']]
        if len(mass) > 0:
            metabolite.notes['mass'] = float(mass)
        metabolite.notes['InChI'] = fields[field_names['InChI']] \
            if len(fields[field_names['InChI']]) > 0 else 'NA'
        metabolite.notes['SMILES'] = fields[field_names['SMILES']] \
            if len(fields[field_names['SMILES']]) > 0 else 'NA'
        metabolite.notes['source'] = fields[field_names['Source']] \
            if len(fields[field_names['Source']]) > 0 else 'NA'
        metabolite.notes['InChIKey'] = fields[field_names['InChIKey']] \
            if len(fields[field_names['InChIKey']]) > 0 else 'NA'
        all_metabolites.append(metabolite)
    LOGGER.info('Finished creating %d Metabolite objects', len(all_metabolites))

    # Download the compartments file.
    compartment_list = _download_metanetx_file('comp_prop.tsv')

    # Map field names to column numbers (MetaNetX may add fields in future).
    field_names = {
        'MNX_ID': 0,
        'Description': 1,
        'Source': 2
    }

    # Add the compartments to the universal model.
    LOGGER.info('Started adding compartments from %d lines in file', len(compartment_list))
    for index in range(len(compartment_list)):
        if len(compartment_list[index]) == 0 or compartment_list[index][0] == '#':
            continue  # Skip empty lines and comment lines
        fields = compartment_list[index].split('\t')
        if len(fields) < len(field_names):
            if verbose:
                warn('Skipped compartment on line {0} with missing fields: {1}'.format(index, compartment_list[index]))
            continue
        universal.compartments[fields[field_names['MNX_ID']]] = fields[field_names['Description']]
    LOGGER.info('Finished adding {0} compartments to universal model'.format(len(universal.compartments)))

    # Download the reactions file.
    reaction_list = _download_metanetx_file('reac_prop.tsv')

    # Map field names to column numbers (hopefully MetaNetX doesn't change this).
    field_names = {
        'MNX_ID': 0,
        'Equation': 1,
        'Description': 2,
        'Balance': 3,
        'EC': 4,
        'Source': 5
    }

    # Accumulate Reaction and Metabolite objects separately because it is faster
    # than adding them one at a time to a model.
    reactions = DictList()
    metabolites = DictList()

    # Create Reaction objects for all of the reactions from the downloaded file.
    LOGGER.info('Started creating Reaction objects from %d lines in file', len(reaction_list))
    for index in range(len(reaction_list)):
        if len(reaction_list[index]) == 0 or reaction_list[index][0] == '#':
            continue  # Skip empty lines and comment lines
        fields = reaction_list[index].split('\t')
        if len(fields) != len(field_names):
            if verbose:
                warn('Skipped reaction on line {0} with missing fields: {1}'.format(index, reaction_list[index]))
            continue

        # Create cobra.core.Reaction from MetaNetX reaction.
        metabolite_info = _parse_metanetx_equation(fields[field_names['Equation']])
        if metabolite_info is None:
            if verbose:
                warn('Could not parse equation for reaction {0} on line {1}: {2}'
                     .format(fields[field_names['MNX_ID']], index, fields[field_names['Equation']]))
            continue
        rxn_mets = dict()
        for metabolite_id in metabolite_info:
            try:
                rxn_mets[metabolites.get_by_id(metabolite_id)] = metabolite_info[metabolite_id]['coefficient']
            except KeyError:
                metabolite = all_metabolites.get_by_id(metabolite_info[metabolite_id]['mnx_id']).copy()
                metabolite.id = metabolite_id
                metabolite.compartment = metabolite_info[metabolite_id]['compartment']
                metabolites.append(metabolite)
                rxn_mets[metabolite] = metabolite_info[metabolite_id]['coefficient']
        reaction = Reaction(id=fields[field_names['MNX_ID']],
                            name=fields[field_names['MNX_ID']],
                            lower_bound=-1000.0,
                            upper_bound=1000.0)
        reaction.add_metabolites(rxn_mets)
        if len(fields[field_names['EC']]) > 0:
            reaction.notes['EC_number'] = fields[field_names['EC']]
        if len(fields[field_names['Source']]) > 1:
            parts = fields[field_names['Source']].split(':')
            if len(parts) == 2:
                reaction.notes['aliases'] = {parts[0]: parts[1]}
            else:
                if verbose:
                    warn('Could not parse source for {0}: {1}'
                         .format(fields[field_names['MNX_ID']], fields[field_names['Source']]))
        reactions.append(reaction)
    LOGGER.info('Finished creating %d Reaction objects', len(reactions))

    # Add the reactions to the universal model.
    universal.add_reactions(reactions)
    LOGGER.info('Finished adding Reaction objects to universal model')

    # If requested, validate the COBRA model.
    if validate:
        warn('Coming soon')

    return universal


def _download_metanetx_file(file_name):
    """ Download and parse a MetaNetX property file.

    Parameters
    ----------
    file_name : str
        Name of property file to download from MetaNetX web site

    Returns
    -------
    list
        List of lines in property file
    """

    LOGGER.info('Started download of %s file', file_name)
    response = requests.get('{0}{1}'.format(metanetx_url, file_name))
    if response.status_code != requests.codes.OK:
        response.raise_for_status()
    LOGGER.info('Finished download of %s file', file_name)
    property_list = response.text.split('\n')
    version = property_list[0].strip('# ')
    if version != metanetx_version:
        warn('MetaNetX version "{0}" in {1} file is not supported version "{2}"'
             .format(version, file_name, metanetx_version))
    return property_list


def _parse_metanetx_equation(equation):
    """ Parse an equation string into a dictionary of metabolite information.

    A MetaNetX equation string has reactants and products separated by an '='
    character and metabolites separate by a '+' character. For example, this
    is the equation for reaction MNXR95362:

    2 MNXM2@MNXD1 + 2 MNXM947@MNXD1 = 1 MNXM4@MNXD1 + 2 MNXM470@MNXD1

    Parameters
    ----------
    equation : str
        Equation string from a MetaNetX reaction

    Returns
    -------
    dict or None
        Dictionary of metabolite information or None if equation string cannot
        be parsed
    """

    # Split the equation string into reactant and product strings.
    parts = equation.split(' = ')
    if len(parts) != 2:
        return None

    # Build a dictionary keyed by metabolite ID with the information needed for
    # setting the metabolites in a cobra.core.Reaction object.
    metabolites = dict()
    reactants = parts[0].split(' + ')
    for r in reactants:
        match = re.search(metabolite_re, r)
        if match is None:
            return None
        met_id = '{0}_{1}'.format(match.group(2), match.group(3))
        metabolites[met_id] = {
            'mnx_id': match.group(2),
            'coefficient': -1.0 * float(match.group(1)),
            'compartment': match.group(3)
        }
    products = parts[1].split(' + ')
    for p in products:
        match = re.search(metabolite_re, p)
        if match is None:
            return None
        met_id = '{0}_{1}'.format(match.group(2), match.group(3))
        metabolites[met_id] = {
            'mnx_id': match.group(2),
            'coefficient': float(match.group(1)),
            'compartment': match.group(3)
        }

    return metabolites



import cobrababel
from cobra.io import write_sbml_model
from cobra.io.sbml3 import validate_sbml_model
from os.path import join
from os import unlink


class TestMetaNetX:
    def test_create_universal(self, test_folder):
        universal = cobrababel.create_metanetx_universal_model()
        assert universal.id == 'metanetx_universal'
        assert len(universal.reactions) >= 42952
        assert len(universal.metabolites) >= 31130
        file_name = join(test_folder, 'metanetx.xml')
        write_sbml_model(universal, file_name)
        model, errors = validate_sbml_model(file_name)
        assert len(errors['other']) == 0
        assert len(errors['SBML errors']) == 0
        assert len(errors['warnings']) == 0
        assert len(errors['validator']) >= 325
        unlink(file_name)



import re

# Regular expression to remove text in parenthesis at end of string.
# @todo Remove leading space?
paren_re = re.compile(r' *\(.*\)$')


class KeggOrganism:
    
    """ A KEGG organism record contains the following fields:

        ID      Each entry is identified by the unique identifier called the T number ('T' followed
                by five-digit number).
        Code    The code for the organism is a three or four character string based on the name
                and is used for identifying genes in complete genomes.
        Name    The scientific name of the organism (common name in parenthesis).
        Taxonomy    The taxonomy of the organism as a string with levels separated by semicolons.
        OtuRep  Flag indicating if the organism is representative of the OTU.

        Each record is a single line with fields separated by tabs (which is different from other record types).
    """
    
    def __init__(self, record=None):
        """ Initialize object.

        Parameters
        ----------
        record : list of str, optional
            List of lines with enzyme record from flat file database
        """
        self.id = None
        self.code = None
        self.name = None
        self.taxonomy = tuple()
        self.otu_representative = 0
        self.search_name = None

        if record is not None:
            self.parse(record)
        return
    
    def as_dict(self):
        """ Return the Organism object as a dictionary.

        Returns
        -------
        dict
            Dictionary representing Organism object
        """

        organism = dict()
        if self.id is not None:
            organism['id'] = self.id
        if self.code is not None:
            organism['code'] = self.code
        if self.name is not None:
            organism['name'] = self.name
        organism['taxonomy'] = self.taxonomy
        organism['otu_representative'] = self.otu_representative
        return organism
    
    def parse(self, record):
        """ Parse a record from the flat file database to complete this object.

        Parameters
        ----------
        record : str
            Line with organism record from flat file database
        """

        # Each record is a single line with the fields separated by tabs. 
        fields = record.split('\t')
        self.id = fields[0]
        self.code = fields[1]
        self.name = fields[2]
        self.taxonomy = tuple(fields[3].split(';'))  # Taxonomy levels are separated by semicolons
        self.otu_representative = int(fields[4])
        self.search_name = re.sub(paren_re, '', self.name)  # Remove text in parenthesis at end of string
        return

    def make_record(self):
        """ Make a record for the flat file database.
        
        Returns
        -------
        str
            Line with organism record
        """
        
        taxonomy = ';'.join(self.taxonomy)
        record = '\t'.join([self.id, self.code, self.name, taxonomy, str(self.otu_representative)])
        return [record]

    # Make these methods properties, then should be able to use with query() method on DictList
    @property
    def is_prokaryote(self):
        """ Check if the organism is a prokaryote.

        Returns
        -------
        bool
            True if the organism is a prokaryote
        """

        return self.taxonomy[0] == 'Prokaryotes'

    @property
    def is_eukaryote(self):
        """ Check if the organism is a eukaryote.

        Returns
        -------
        bool
            True if the organism is a eukaryote
        """
        
        return self.taxonomy[0] == 'Eukaryotes'

    @property
    def is_bacteria(self):
        """ Check if the organism is a bacteria.

        Returns
        -------
        bool
            True if the organism is a bacteria
        """
        
        return self.taxonomy[1] == 'Bacteria'

    @property
    def is_archaea(self):
        """ Check if the organism is an archaea.

        Returns
        -------
        bool
           True if the organism is an archaea
        """
        
        return self.taxonomy[1] == 'Archaea'

    @property
    def is_representative(self):
        """ Check if the organism is a representative of the OTU.

        Returns
        -------
        bool
            True if organism is an OTU representative
        """

        return self.otu_representative



from six import iteritems
import json
import pandas as pd
from tabulate import tabulate
from cobra.core import DictList
from warnings import warn
from numpy import isclose

from .util import format_long_string

"""
Notes
-----
    For ModelSEED models, I don't understand the boundary reactions rxn13782_c, rxn13783_c,
    and rxn13784_c which I think are sink reactions for protein biosynthesis, DNA replication,
    and RNA transcrption.  The compounds show up as being consumed but are produced by the
    biomass reaction.
"""

# Header lines for tabulated output
reaction_header = ['ID', 'NAME', 'REACTION']
metabolite_header = ['ID', 'NAME']
gene_header = ['ID', 'NAME']
difference_header = ['ID', 'FIRST', 'SECOND']


def compare_models(model1, model2, details=None, boundary=False):
    """ Compare two models and report differences.

    For a useful comparison, the models must use the same ID types. For example,
    comparing models that use ModelSEED IDs is valid but comparing a model that
    uses ModelSEED IDs with a model that uses BiGG IDs does not work.

    Parameters
    ----------
    model1 : cobra.core.Model
        First model to analyze
    model2 : cobra.core.Model
        Second model to analyze
    details : set, optional
        When specified, print details on given types of differences (see other compare functions)
    boundary : bool, optional
        When true, print info about boundary reactions
    """

    # Compare reactions, metabolites, and genes.
    compare_reactions(model1.reactions, model2.reactions, details=details, id1=model1.id, id2=model2.id)
    compare_metabolites(model1.metabolites, model2.metabolites, details=details, id1=model1.id, id2=model2.id)
    compare_genes(model1.genes, model2.genes, details=details, id1=model1.id, id2=model2.id)

    # See about system boundary reactions.
    if boundary:
        # Get the list of system boundary reactions from first model.
        model1_boundary = model1.reactions.query(lambda x: x, 'boundary')
        print('{0} reactions are system boundary reactions in {1}'.format(len(model1_boundary), model1.id))

        # Get the list of system boundary reactions from second model.
        model2_boundary = model2.reactions.query(lambda x: x, 'boundary')
        print('{0} reactions are system boundary reactions in {1}'.format(len(model2_boundary), model2.id))

    return


def compare_reactions(reaction1, reaction2, details=None, id1='first', id2='second'):
    """ Compare two lists of cobra.core.Reaction objects and report differences.

    To determine if two reactions are the same, the function compares the following 
    attributes: (1) ID {'reaction_id'}, (2) name {'reaction_name'}, (3) bounds
    {'reaction_bounds'}, (4) definition {'reaction_definition'}, (5) gene reaction 
    rule {'reaction_gpr'}. Include the value in {} in the details parameter to
    display the details of reactions where the values are different.
    
    Parameters
    ----------
    reaction1 : cobra.core.DictList
        First list of cobra.core.Reaction objects to analyze
    reaction2 : cobra.core.DictList
        Second list of cobra.core.Reaction objects to analyze
    details : set, optional
        When specified, print details on given types of differences
    id1 : str, optional
        ID for labeling first list of reactions
    id2 : str, optional
        ID for labeling second list of reactions
    """

    if details is None:
        details = set()

    print('REACTIONS\n' + '---------')
    print('{0} reactions in {1}'.format(len(reaction1), id1))
    print('{0} reactions in {1}\n'.format(len(reaction2), id2))

    # See if reactions from first model are in the second model.
    num_matched = 0
    reaction_only_in_one = DictList()
    different_name = DictList()
    different_bounds = DictList()
    different_definition = DictList()
    different_genes = DictList()
    for r1 in reaction1:
        try:
            r2 = reaction2.get_by_id(r1.id)
            num_matched += 1
            if r1.name != r2.name:
                different_name.append(r1)
            if r1.bounds != r2.bounds:
                different_bounds.append(r1)
            if r1.reaction != r2.reaction:
                different = False
                for met, coefficient in iteritems(r1.metabolites):
                    if not isclose(r2.get_coefficient(met.id), coefficient):
                        different = True
                if different:
                    different_definition.append(r1)
            if r1.gene_reaction_rule != r2.gene_reaction_rule:
                different_genes.append(r1)
        except KeyError:
            reaction_only_in_one.append(r1)
    print('{0} reactions in {1} and {2}'.format(num_matched, id1, id2))
    print('{0} reactions only in {1}\n'.format(len(reaction_only_in_one), id1))

    # If requested, show the details on reactions only in the first model.
    if 'reaction_id' in details and len(reaction_only_in_one) > 0:
        reaction_only_in_one.sort(key=lambda x: x.id)
        output = [[rxn.id, format_long_string(rxn.name, 20), rxn.reaction]
                  for rxn in reaction_only_in_one]
        print(tabulate(output, tablefmt='simple', headers=reaction_header) + '\n')

    # See if reactions from second model are in the first model.
    num_matched = 0
    reaction_only_in_two = DictList()
    for r2 in reaction2:
        if reaction1.has_id(r2.id):
            num_matched += 1
        else:
            reaction_only_in_two.append(r2)
    print('{0} reactions in both {1} and {2}'.format(num_matched, id1, id2))
    print('{0} reactions only in {1}\n'.format(len(reaction_only_in_two), id2))

    # If requested, show the details on reactions only in the second model.
    if 'reaction_id' in details and len(reaction_only_in_two) > 0:
        reaction_only_in_two.sort(key=lambda x: x.id)
        output = [[rxn.id, format_long_string(rxn.name, 20), rxn.reaction]
                  for rxn in reaction_only_in_two]
        print('\n' + tabulate(output, tablefmt='simple', headers=reaction_header) + '\n')

    # Display details on reaction attribute differences.
    print('{0} reactions with different names'.format(len(different_name)))
    if 'reaction_name' in details and len(different_name) > 0:
        different_name.sort(key=lambda x: x.id)
        output = [[rxn.id, rxn.name, reaction2.get_by_id(rxn.id).name]
                  for rxn in different_name]
        print('\n' + tabulate(output, tablefmt='simple', headers=difference_header) + '\n')
    print('{0} reactions with different bounds'.format(len(different_bounds)))
    if 'reaction_bounds' in details and len(different_bounds) > 0:
        different_bounds.sort(key=lambda x: x.id)
        output = [[rxn.id, rxn.bounds, reaction2.get_by_id(rxn.id).bounds]
                  for rxn in different_bounds]
        print('\n' + tabulate(output, tablefmt='simple', headers=difference_header) + '\n')
    print('{0} reactions with different definitions'.format(len(different_definition)))
    if 'reaction_definition' in details and len(different_definition) > 0:
        different_definition.sort(key=lambda x: x.id)
        output = [[rxn.id, rxn.reaction, reaction2.get_by_id(rxn.id).reaction]
                  for rxn in different_definition]
        print('\n' + tabulate(output, tablefmt='simple', headers=difference_header) + '\n')
    print('{0} reactions with different genes'.format(len(different_genes)))
    if 'reaction_gpr' in details and len(different_genes) > 0:
        different_genes.sort(key=lambda x: x.id)
        output = [[rxn.id, rxn.gene_reaction_rule, reaction2.get_by_id(rxn.id).gene_reaction_rule]
                  for rxn in different_genes]
        print('\n' + tabulate(output, tablefmt='simple', headers=difference_header) + '\n')

    return


def compare_metabolites(metabolite1, metabolite2, details=None, id1='first', id2='second'):
    """ Compare two lists of cobra.core.Metabolite objects and report differences.

    To determine if two metabolites are the same, the function compares the following 
    attributes: (1) ID {'metabolite_id'}, (2) name {'metabolite_name'}, (3) formula
    {'metabolite_formula'}, (4) charge {'metabolite_charge'}, (5) compartment 
    {'metabolite_compartment'}. Include the value in {} in the details parameter to 
    display the details of metabolites where the values are different.
    
    Parameters
    ----------
    metabolite1 : cobra.core.DictList
        First list of cobra.core.Metabolite objects to analyze
    metabolite2 : cobra.core.DictList
        Second list of cobra.core.Metabolite objects to analyze
    details : set, optional
        When specified, print details on given types of differences
    id1 : str, optional
        ID for labeling first list of metabolites
    id2 : str, optional
        ID for labeling second list of metabolites
    """

    if details is None:
        details = set()

    print('\nMETABOLITES\n' + '-----------')
    print('{0} metabolites in {1}'.format(len(metabolite1), id1))
    print('{0} metabolites in {1}\n'.format(len(metabolite2), id2))

    # See if metabolites from first model are in the second model.
    num_matched = 0
    metabolite_only_in_one = DictList()
    different_name = DictList()
    different_formula = DictList()
    different_charge = DictList()
    different_compartment = DictList()
    for m1 in metabolite1:
        try:
            m2 = metabolite2.get_by_id(m1.id)
            num_matched += 1
            if m1.name != m2.name:
                different_name.append(m1)
            if m1.formula != m2.formula:
                different_formula.append(m1)
            if m1.charge != m2.charge:
                different_charge.append(m1)
            if m1.compartment != m2.compartment:
                different_compartment.append(m1)
        except KeyError:
            metabolite_only_in_one.append(m1)
    print('{0} metabolites in both {1} and {2}'.format(num_matched, id1, id2))
    print('{0} metabolites only in {1}\n'.format(len(metabolite_only_in_one), id1))
    if 'metabolite_id' in details and len(metabolite_only_in_one) > 0:
        metabolite_only_in_one.sort(key=lambda x: x.id)
        output = [[met.id, format_long_string(met.name, 70)] for met in metabolite_only_in_one]
        print('\n' + tabulate(output, tablefmt='simple', headers=metabolite_header) + '\n')

    # See if metabolites from second model are in the first model.
    num_matched = 0
    metabolite_only_in_two = DictList()
    for m2 in metabolite2:
        if metabolite1.has_id(m2.id):
            num_matched += 1
        else:
            metabolite_only_in_two.append(m2)
    print('{0} metabolites in both {1} and {2}'.format(num_matched, id1, id2))
    print('{0} metabolites only in {1}\n'.format(len(metabolite_only_in_two), id2))
    if 'metabolite_id' in details and len(metabolite_only_in_two) > 0:
        metabolite_only_in_two.sort(key=lambda x: x.id)
        output = [[met.id, format_long_string(met.name, 70)] for met in metabolite_only_in_two]
        print('\n' + tabulate(output, tablefmt='simple', headers=metabolite_header) + '\n')

    # Display details on metabolite attribute differences.
    print('{0} metabolites with different names'.format(len(different_name)))
    if 'metabolite_name' in details and len(different_name) > 0:
        different_name.sort(key=lambda x: x.id)
        output = [[met.id, met.name, metabolite2.get_by_id(met.id).name]
                  for met in different_name]
        print('\n' + tabulate(output, tablefmt='simple', headers=difference_header) + '\n')
    print('{0} metabolites with different formulas'.format(len(different_formula)))
    if 'metabolite_formula' in details and len(different_formula) > 0:
        different_formula.sort(key=lambda x: x.id)
        output = [[met.id, met.formula, metabolite2.get_by_id(met.id).formula]
                  for met in different_formula]
        print('\n' + tabulate(output, tablefmt='simple', headers=difference_header) + '\n')
    print('{0} metabolites with different charges'.format(len(different_charge)))
    if 'metabolite_charge' in details and len(different_charge) > 0:
        different_charge.sort(key=lambda x: x.id)
        output = [[met.id, met.charge, metabolite2.get_by_id(met.id).charge]
                  for met in different_charge]
        print('\n' + tabulate(output, tablefmt='simple', headers=difference_header) + '\n')
    print('{0} metabolites with different compartments'.format(len(different_compartment)))
    if 'metabolite_compartment' in details and len(different_compartment) > 0:
        different_compartment.sort(key=lambda x: x.id)
        output = [[met.id, met.compartment, metabolite2.get_by_id(met.id).compartment]
                  for met in different_compartment]
        print('\n' + tabulate(output, tablefmt='simple', headers=difference_header) + '\n')

    return


def compare_genes(gene1, gene2, details=None, id1='first', id2='second'):
    """ Compare two lists of cobra.core.Gene objects and report differences.

    To determine if two genes are the same, the function compares the following 
    attributes: (1) ID {'gene_id'}, (2) name {'gene_name'}. Include the value 
    in {} in the details parameter to display the details of genes where the 
    values are different.
    
    Parameters
    ----------
    gene1 : cobra.core.DictList
        First list of cobra.core.Gene objects to analyze
    gene2 : cobra.core.DictList
        Second list of cobra.core.Gene objects to analyze
    details : set, optional
        When specified, print details on given types of differences
    id1 : str, optional
        ID for labeling first list of genes
    id2 : str, optional
        ID for labeling second list of genes
    """

    if details is None:
        details = set()

    print('\nGENES\n' + '------')
    print('{0} genes in {1}'.format(len(gene1), id1))
    print('{0} genes in {1}\n'.format(len(gene2), id2))

    # See if genes from first list are in the second list.
    num_matched = 0
    gene_only_in_one = DictList()
    different_name = DictList()
    for g1 in gene1:
        try:
            g2 = gene2.get_by_id(g1.id)
            num_matched += 1
            if g1.name.lower() != g2.name.lower():
                different_name.append(g1)
        except KeyError:
            gene_only_in_one.append(g1)
    print('{0} genes in both {1} and {2}'.format(num_matched, id1, id2))
    print('{0} genes only in {1}\n'.format(len(gene_only_in_one), id1))
    if 'gene_id' in details and len(gene_only_in_one) > 0:
        gene_only_in_one.sort(key=lambda x: x.id)
        output = [[gene.id, format_long_string(gene.name, 90)] for gene in gene_only_in_one]
        print('\n' + tabulate(output, tablefmt='simple', headers=gene_header) + '\n')

    # See if genes from second list are in the first list.
    num_matched = 0
    gene_only_in_two = DictList()
    for g2 in gene2:
        if gene1.has_id(g2.id):
            num_matched += 1
        else:
            gene_only_in_two.append(g2)
    print('{0} genes in both {1} and {2}'.format(num_matched, id1, id2))
    print('{0} genes only in {1}\n'.format(len(gene_only_in_two), id2))
    if 'gene_id' in details and len(gene_only_in_two) > 0:
        gene_only_in_two.sort(key=lambda x: x.id)
        output = [[gene.id, format_long_string(gene.name, 90)] for gene in gene_only_in_two]
        print('\n' + tabulate(output, tablefmt='simple', headers=gene_header) + '\n')

    # Display details on gene attribute differences.
    print('{0} genes with different names'.format(len(different_name)))
    if 'gene_name' in details and len(different_name) > 0:
        different_name.sort(key=lambda x: x.id)
        output = [[gene.id, gene.name, gene2.get_by_id(gene.id).name]
                  for gene in different_name]
        print('\n' + tabulate(output, tablefmt='simple', headers=difference_header) + '\n')

    return



import requests
from warnings import warn

from .KeggReaction import KeggReaction
from .KeggEnzyme import KeggEnzyme

# Base URL for KEGG website
kegg_url = 'http://rest.kegg.jp/'


class QueryError(Exception):
    """ Exception raised when there is an error with a KEGG query """
    pass


def _kegg_request(url, result):
    """ Send a request to the KEGG web service.

    Parameters
    ----------
    url : str
        KEGG API request URL
    result : list of str
        List of lines from query result

    Returns
    -------
    list of str
        Updated list of lines with results from this query
    """

    # Send the request and receive the response from the web service.
    response = requests.get(url)
    if response.status_code != requests.codes.OK:
        response.raise_for_status()

    # When the status code indicates a failure, raise an exception with the details.
    # @todo will this even run?
    if response.status_code != 200:
        if response.status_code == 400:
            reason = 'bad request (syntax error, wrong database name, etc.)'
        elif response.status_code == 404:
            reason = 'not found'
        else:
            reason = 'unknown ({0})'.format(response.status_code)
        raise QueryError('Query {0} failed: {1}'.format(url, reason))

    # Add the response to the current results.
    result.extend(response.text.split('\n')[:-1])  # Remove the empty last line
    return result


def get_kegg_records(db_name, id_list, option=None):
    """ Get records from a KEGG database.

    Parameters
    ----------
    db_name : str
        Name of database to get_kegg_records records from
    id_list : list of str
        List of IDs of records to get_kegg_records
    option : str
        Option value for request

    Returns
    -------
    list of str
         List of lines from records for specified IDs
    """

    # The web service limits the size of the query to 10 items.
    increment = 10
    start = 0
    if len(id_list) > increment:
        end = start + increment
    else:
        end = len(id_list)
    counter = len(id_list)

    # Run through the list of IDs until all have been processed.
    result = list()
    while counter > 0:
        # Build the query string.
        query = ''
        for index in range(start, end):
            query += db_name + ':' + id_list[index] + '+'

        # Adjust for the next query.
        start += increment
        end += increment
        if end > len(id_list):
            end = len(id_list)
        counter -= increment

        # Send the request.
        url = '{0}get/{1}'.format(kegg_url, query[:-1])  # Take off trailing +
        if option is not None:
            url = url + '/' + option
        result = _kegg_request(url, result)

    return result


def list_kegg_ids(db_name):
    """ Return a list of entry identifiers and associated definitions for a given database.

    Parameters
    ----------
    db_name : str
        Name of database to query

    Returns
    -------
    list of str
         List of lines from query result
    """

    # Retrieve the list.
    return _kegg_request('{0}list/{1}'.format(kegg_url, db_name), list())


def get_kegg_metabolites(id_list):

    warn('Getting metabolites (or compounds) from KEGG is not supported yet')
    return


def get_kegg_reactions(id_list):
    """ Get reactions from the KEGG reaction database.

    Parameters
    ----------
    id_list : list of str
        List of KEGG reaction IDs

    Returns
    -------
    list of KeggReaction objects
         Objects for returned reactions
    """

    # Get the records from the reaction database.
    record_list = get_kegg_records('rn', id_list)

    # Convert the returned reaction records to KeggReaction objects.
    reaction_list = list()
    start = 0
    for index in range(len(record_list)):
        if record_list[index] == '///':
            reaction_list.append(KeggReaction(record_list[start:index + 1]))
            start = index + 1

    return reaction_list


def get_kegg_enzymes(id_list):
    """ Get enzymes from the enzyme database.

    Parameters
    ----------
    id_list : list of str
        List of KEGG enzyme IDs

    Returns
    -------
    list of KEGGEnyzme objects
         Objects for returned enzymes
    """

    # Get the records from the enzyme database.
    record_list = get_kegg_records('ec', id_list)

    # Convert the returned enzyme records to KeggEnzyme objects.
    enzyme_list = list()
    start = 0
    for index in range(len(record_list)):
        if record_list[index] == '///':
            enzyme_list.append(KeggEnzyme(record_list[start:index + 1]))
            start = index + 1

    return enzyme_list


def get_kegg_amino_acid_seq(code, gene_list):
    """ Get amino acid sequences for a list of genes from an organism.

    Parameters
    ----------
    code : str
        Organism code (3 or 4 character string)
    gene_list : list of str
        List of KEGG gene IDs in organism

    Returns
    -------
    list of str
        List of lines in FASTA format with amino acid sequence for specified genes
    """

    # @todo Maybe the sequences should be returned in a better data structure
    return get_kegg_records(code, gene_list, option='aaseq')


def get_kegg_dna_seq(code, gene_list):
    """ Get DNA sequences for a list of genes from an organism.

    Parameters
    ----------
    code : str
        Organism code (3 or 4 character string)
    gene_list : list of str
        List of KEGG gene IDs in organism

    Returns
    -------
    list of str
        List of lines in FASTA format with DNA sequence for specified genes
    """

    return get_kegg_records(code, gene_list, option='ntseq')



import requests
from warnings import warn
from tempfile import gettempdir
from os.path import join
from os import unlink
import zipfile
import io

from cobra.io import read_sbml_model, load_matlab_model

# Base URL for Virtual Metabolic Human website
vmh_url = 'https://webdav-r3lab.uni.lu/public/msp/'

# File name of current Recon2 model.
recon2_file_name = 'Recon2.v04.mat'


def create_cobra_model_from_agora_model(agora_name, validate=False):
    """ Create a COBRA model from an AGORA model.

    Parameters
    ----------
    agora_name: str
        Name of AGORA model
    validate : bool, optional
        When True, perform validity checks on COBRA model

    Returns
    -------
    cobra.Model
        COBRA model created from SBML representation of AGORA model
    """

    # Download the SBML file.
    response = requests.get('{0}AGORA/sbml/{1}.xml'.format(vmh_url, agora_name))
    if response.status_code != requests.codes.OK:
        response.raise_for_status()

    # Convert to a cobra.Model object.
    with io.BytesIO(response.content) as f:
        model = read_sbml_model(f)

    # If requested, validate the COBRA model.
    if validate:
        warn('Coming soon')

    return model


def create_cobra_model_from_vmh_recon2(validate=False):
    """ Create a COBRA model from the current Recon2 model.

    Parameters
    ----------
    validate : bool, optional
        When True, perform validity checks on COBRA model

    Returns
    -------
    cobra.Model
        COBRA model created from Matlab representation of Recon2 model
    """

    # Download the zip file that contains the Matlab file and extract it.
    response = requests.get('{0}{1}_.zip'.format(vmh_url, recon2_file_name), stream=True)
    if response.status_code != requests.codes.OK:
        response.raise_for_status()
    zipfile.ZipFile(io.BytesIO(response.content), 'r').extract(recon2_file_name, gettempdir())

    # Convert to a cobra.Model object.
    temp_file = join(gettempdir(), recon2_file_name)
    model = load_matlab_model(temp_file)
    unlink(temp_file)

    # If requested, validate the COBRA model.
    if validate:
        warn('Coming soon')

    return model



import pytest
import cobrababel
from cobra import Model


class TestBigg:
    def test_model_list(self):
        model_list = cobrababel.get_bigg_model_list()
        assert len(model_list) >= 80
        assert 'bigg_id' in model_list[0]
        assert 'organism' in model_list[0]
        assert 'metabolite_count' in model_list[0]
        assert 'reaction_count' in model_list[0]
        assert 'gene_count' in model_list[0]

    def test_create_model(self):
        model = cobrababel.create_cobra_model_from_bigg_model('iAF1260')
        assert len(model.metabolites) == 1668
        assert len(model.reactions) == 2382
        assert len(model.genes) == 1261
        solution = model.optimize()
        assert solution.f == pytest.approx(0.736701)

    def test_add_metabolite(self):
        model = Model('bigg_test')
        cobrababel.add_bigg_metabolites([cobrababel.get_bigg_metabolite('h2o_c', 'iAF1260')], model)
        assert model.metabolites[0].id == 'h2o_c'
        assert model.metabolites[0].name == 'H2O'
        assert model.metabolites[0].compartment == 'c'
        cobrababel.add_bigg_metabolites([cobrababel.get_bigg_metabolite('10fthf')], model)
        assert model.metabolites[1].id == '10fthf_c'
        assert model.metabolites[1].name == '10-Formyltetrahydrofolate'
        assert model.metabolites[1].compartment == 'c'
        assert model.compartments['c'] == 'cytosol'

    def test_add_reaction(self):
        model = Model('bigg_test')
        cobrababel.add_bigg_metabolites([cobrababel.get_bigg_metabolite('cynt_p', 'iAF1260'),
                                         cobrababel.get_bigg_metabolite('h_p', 'iAF1260'),
                                         cobrababel.get_bigg_metabolite('cynt_c', 'iAF1260'),
                                         cobrababel.get_bigg_metabolite('h_c', 'iAF1260')], model)
        cobrababel.add_bigg_reactions([cobrababel.get_bigg_reaction('CYNTt2pp', 'iAF1260')], model)
        assert model.reactions[0].id == 'CYNTt2pp'
        assert model.reactions[0].name == 'Cyanate transport via proton symport (periplasm)'
        assert model.reactions[0].reaction == 'cynt_p + h_p --> cynt_c + h_c'
        assert len(model.reactions[0].metabolites) == 4
        assert len(model.compartments) == 2



from os.path import splitext
from warnings import warn
import re

from cobra.io import load_matlab_model, load_json_model, read_sbml_model, save_json_model
from cobra.core import Model

# Regular expression for compartment suffix on ModelSEED IDs
modelseed_suffix_re = re.compile(r'_([ce])$')

# Regular expression for compartment suffix on BiGG IDs
bigg_suffix_re = re.compile(r'\[([cefghilmnpsruvx])\]$')


def _id_type(object_id):
    """ Figure out the ID type for an object ID.

    Parameters
    ----------
    object_id : str

    Returns
    -------
    str
        ID type ('bigg', 'modelseed', or 'unknown')
    """

    if re.search(bigg_suffix_re, object_id) is not None:
        return 'bigg'
    if re.search(modelseed_suffix_re, object_id) is not None:
        return 'modelseed'
    return 'unknown'


def load_model_from_file(filename):
    """ Load a model from a file based on the extension of the file name.

    Parameters
    ----------
    filename : str
        Path to model file

    Returns
    -------
    cobra.Model
        Model object loaded from file

    Raises
    ------
    Exception
        If model file extension is not supported.
    """

    (root, ext) = splitext(filename)
    if ext == '.mat':
        model = load_matlab_model(filename)
    elif ext == '.xml' or ext == '.sbml':
        model = read_sbml_model(filename)
    elif ext == '.json':
        model = load_json_model(filename)
    else:
        raise IOError('Model file extension not supported for {0}'.format(filename))

    return model


def create_universal_model_from_source(source_models, file_name=None, validate=False):
    """ Create an universal model from a list of source models.

    Parameters
    ----------
    source_models : list of str
        List of path names to source model files
    file_name : str, optional
        Path to file for saving universal COBRA model in JSON format
    validate : bool, optional
        When True, perform validity checks on universal COBRA model

    Returns
    -------
    cobra.Model
        COBRA model object with universal reactions and metabolites
    """

    # Create a new COBRApy Model object.
    universal = Model('universal', name='Universal')
    universal.notes['sources'] = list()

    # Add all of the source models to the universal model.
    for model_filename in source_models:
        # Load the model from a file.
        model = load_model_from_file(model_filename)

        # All metabolites need to have a compartment suffix.
        for metabolite in model.metabolites:
            metabolite.notes['type'] = _id_type(metabolite.id)
        unknown = model.metabolites.query(lambda x: 'unknown' in x['type'], 'notes')
        if len(unknown) > 0:
            raise Exception('Unknown compartment suffixes found in metabolites for {0}'.format(model_filename))

        # Remove any duplicate reactions from model which will leave a model with just the new reactions.
        duplicates = model.reactions.query(lambda x: universal.reactions.has_id(x), 'id')
        model.remove_reactions(duplicates, remove_orphans=True)

        # Remove objectives and genes from model.
        model.objective = {}
        for reaction in model.reactions:
            reaction.gene_reaction_rule = ''

        # Add new metabolites and reactions from the model to the universal model.
        universal += model

        # Add attributes not automatically added above.
        universal.notes['sources'].append({'id': model.id, 'filename': model_filename})
        for compartment in model.compartments:
            if compartment in universal.compartments:
                if model.compartments[compartment] != universal.compartments[compartment]:
                    warn('Model {0} has a different meaning for compartment {1} {2}'
                         .format(model.id, compartment, model.compartments[compartment]))
            else:
                universal.compartments[compartment] = model.compartments[compartment]

    # If requested, validate the COBRA model.
    if validate:
        warn('Coming soon')

    # If requested, save the COBRA model.
    if file_name is not None:
        save_json_model(universal, file_name)

    return universal



import re
from warnings import warn

from .KeggDatabase import KeggDatabase
from .KeggEnzyme import KeggEnzyme


class KeggEnzymeDatabase(KeggDatabase):
    """ Manage a KEGG enzyme flat file database. """

    def __init__(self, filename):
        """ Initialize object.
        
        Parameters
        ----------
        filename : str
            Path to enzyme database file
        """

        super(KeggEnzymeDatabase, self).__init__(filename)
        return
    
    def load(self):
        """ Load the enzyme database from a flat file. """
        
        with open(self.filename, 'r') as handle:
            enzymes = [KeggEnzyme(record) for record in self.get_record(handle)]
        self.records += enzymes
        return
    
    def store(self, filename=None):
        """ Save the enzyme database to a flat file.
        
        Parameters
        ----------
        filename : str, optional
            Path to enzyme database file
        """
        
        if filename is None:
            filename = self.filename
        
        # Convert all of the KeggEnzyme objects to flat file database records
        # and write the records to the file.
        self.records.sort()
        with open(filename, 'w') as handle:
            for index in range(len(self.records)):
                for line in self.records[index].make_record():
                    handle.write(line+'\n')
        return

    def update(self, enzyme):
        """ Update an enzyme in the database (add new or replace existing enzyme).
        
        Parameters
        ----------
        enzyme : KeggEnzyme
            Enzyme to add or replace
        """

        # Replace the current KeggEnzyme object if it already exists in the database.
        if self.records.has_id(enzyme.id):
            self.records._replace_on_id(enzyme)
            return
            
        # Add the new KeggEnzyme object to the database.
        self.records += [enzyme]
        return

    def validate(self):
        """ Validate that obsolete enzymes are linked to valid enzymes in the database.

        Returns
        -------
        dict
            Dictionary with counts and list of IDs of missing enzymes
        """

        # Keep track of what is found.
        obsolete = dict()
        obsolete['num_obsolete'] = 0
        obsolete['num_transferred'] = 0
        obsolete['num_deleted'] = 0
        obsolete['missing_enzymes'] = list()
        ec_number_re = re.compile(r'([\d\-]+\.[\d\-]+\.[\d\-]+\.[\d\-]+)')

        # Check each enzyme in the database to see it is obsolete.
        for enzyme in self.records:
            # enzyme = self.records.get_by_id(id)
            if enzyme.obsolete:
                obsolete['num_obsolete'] += 1
                if len(enzyme.name) > 1:
                    warn('Obsolete enzyme {0} has more than one name: {1}'.format(id, enzyme.name))

                # When an obsolete enzyme has been transferred to another enzyme, the name field
                # has the reference to the replacement enzyme.
                if enzyme.name[0].startswith('Transferred to '):
                    obsolete['num_transferred'] += 1
                    matches = re.findall(ec_number_re, enzyme.name[0])
                    for ec_number in matches:
                        if not self.records.has_id(ec_number):
                            obsolete['missing_enzymes'].append(ec_number)

                # When an obsolete enzyme has been deleted, the comment field may have a
                # reference to a replacement enzyme.
                elif enzyme.comment[0].startswith('Deleted entry: '):
                    obsolete['num_deleted'] += 1
                    matches = re.findall(ec_number_re, enzyme.comment[0])
                    for ec_number in matches:
                        if not self.records.has_id(ec_number):
                            obsolete['missing_enzymes'].append(ec_number)
                else:
                    warn('Obsolete enzyme {0} has not been transferred or deleted'.format(id))

        return obsolete



from setuptools import setup, find_packages

requirements = [
    'cobra>=0.5.4',
    'six',
    'pandas>=0.18.0',
    'fuzzywuzzy>=0.10.0',
    'requests',
    'tabulate',
    'numpy>=1.6'
]

try:
    with open('README.rst') as handle:
        description = handle.read()
except:
    description = ''

setup(
    name='cobrababel',
    version='0.1.3',
    packages=find_packages(),
    setup_requires=[],
    install_requires=requirements,
    tests_require=['pytest', 'lxml'],
    package_data={},
    author='Michael Mundy, Helena Mendes-Soares, Nicholas Chia',
    author_email='mundy.michael@mayo.edu',
    description='CobraBabel: COBRA Model Translator',
    long_description=description,
    license='BSD',
    keywords='metabolism biology optimization flux balance analysis fba',
    url='https://github.com/mmundy42/cobrababel',
    download_url='https://pypi.python.org/pypi/cobrababel',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ],
    platforms='GNU/Linux, Mac OS X >= 10.7, Microsoft Windows >= 7'
)



import cobrababel
import pytest
from requests import HTTPError


class TestKEGG:
    def test_list_ids(self):
        id_list = cobrababel.list_kegg_ids('rn')
        assert len(id_list) >= 10550
        assert id_list[0].startswith('rn:R00001')

    def test_list_id_bad_name(self):
        with pytest.raises(HTTPError):
            cobrababel.list_kegg_ids('xx')

    def test_get_reactions(self):
        rxn_list = cobrababel.get_kegg_reactions(['R00001', 'R00013', 'R01175'])
        assert len(rxn_list) == 3
        assert rxn_list[0].id == 'R00001'
        assert rxn_list[1].id == 'R00013'
        assert rxn_list[2].id == 'R01175'

    def test_get_enzymes(self):
        enzyme_list = cobrababel.get_kegg_enzymes(['1.1.1.1', '1.1.1.101', '1.5.3.7'])
        assert len(enzyme_list) == 3
        assert enzyme_list[0].id == '1.1.1.1'
        assert enzyme_list[1].id == '1.1.1.101'
        assert enzyme_list[2].id == '1.5.3.7'

    def test_get_amino_seq(self):
        aa_seq = cobrababel.get_kegg_amino_acid_seq('T00122', ['BT_0001', 'BT_0009', 'BT_0015'])
        assert len(aa_seq) == 22
        assert aa_seq[0].startswith('>bth:BT_0001')
        assert aa_seq[1].startswith('MVSTS')
        assert aa_seq[5].startswith('>bth:BT_0009')
        assert aa_seq[6].startswith('MATIR')
        assert aa_seq[13].startswith('>bth:BT_0015')
        assert aa_seq[14].startswith('MKSIL')

    def test_get_dna_seq(self):
        dna_seq = cobrababel.get_kegg_dna_seq('T00122', ['BT_0002', 'BT_0033', 'BT_0042'])
        assert len(dna_seq) == 45
        assert dna_seq[0].startswith('>bth:BT_0002')
        assert dna_seq[1].startswith('atgatgaa')
        assert dna_seq[18].startswith('>bth:BT_0033')
        assert dna_seq[19].startswith('atgatta')
        assert dna_seq[24].startswith('>bth:BT_0042')
        assert dna_seq[25].startswith('atgaaaa')



from .KeggDatabase import KeggDatabase
from .KeggReaction import KeggReaction

class KeggReactionDatabase(KeggDatabase):
    """ Manage a KEGG reaction flat file database. """

    def __init__(self, filename):
        """ Initialize object.

        Parameters
        ----------
        filename : str
            Path to reaction database file
\        """

        super(KeggReactionDatabase, self).__init__(filename)
        return
    
    def load(self):
        """ Load the reaction database from a flat file. """
        
        with open(self.filename, 'r') as handle:
            reactions = [KeggReaction(record) for record in self.get_record(handle)]
        self.records += reactions
        return
    
    def store(self, filename=None):
        """ Save the reaction database to a flat file.

        Parameters
        ----------
        filename : str, optional
            Path to reaction database file
\       """
        
        if filename is None:
            filename = self.filename
            
        # Convert all of the KeggReaction objects to flat file database records
        # and write the records to the file.    
        self.records.sort()
        with open(filename, 'w') as handle:
            for index in range(len(self.records)):
                for line in self.records[index].make_record():
                    handle.write(line + '\n')
        return

    def update(self, reaction):
        """ Update a reaction in the database (add new or replace existing reaction).

        Parameters
        ----------
        reaction : KeggReaction
            Reaction to add or replace
        """

        # Replace the current KeggReaction object if it already exists in the database.
        if self.records.has_id(reaction.id):
            self.records._replace_on_id(reaction)
            return

        # Add the new KeggReaction object to the database.
        self.records += [reaction]
        return



from fuzzywuzzy import fuzz
from pandas import DataFrame, Series

from .KeggDatabase import KeggDatabase, DatabaseError
from .KeggOrganism import KeggOrganism
from .kegg import list_kegg_ids

# Cutoff for identifying a good match when comparing names.
MATCH_CUTOFF = 97

# Column names for reporting results of setting OTU representatives.
otu_report_columns = ['OTU_REP_NAME', 'OTU_REP_ID', 'NAME', 'ID', 'RATIO',
                      'PARTIAL_RATIO', 'TOKEN_SET_RATIO', 'RESULT']


class KeggOrganismDatabase(KeggDatabase):
    """ Manage a KEGG organism flat file database. """

    def __init__(self, filename):
        """ Initialize object.
        
        Parameters
        ----------
        filename : str
            Path to organism database file
        """

        super(KeggOrganismDatabase, self).__init__(filename)
        self.code_to_id = dict()  # Keyed by organism code
        return

    def load(self):
        """ Load the organism database from a flat file. """

        with open(self.filename, 'r') as handle:
            # Organism database is different and each record is one tab-delimited line.
            organisms = [KeggOrganism(record) for record in handle]
        self.records += organisms

        # Build a dictionary to lookup organisms by code.
        for index in range(len(self.records)):
            self.code_to_id[self.records[index].code] = self.records[index].id
        return

    def download(self):
        """ Download the organism database from KEGG web service.

            The only way to get_kegg_records taxonomic info for organisms is with the "list" operation
            from the KEGG web service.
        """

        # Make sure the database is empty.
        if len(self.records) > 0:
            raise DatabaseError('Organism database must be empty before downloading from web service')

        # Get the current organism database from KEGG.
        organism_list = list_kegg_ids('organism')

        # Add all of the organisms to the database.
        organisms = list()
        for index in range(len(organism_list)):
            record = organism_list[index]
            record += '\t0'  # Mark as not a representative by default
            organisms.append(KeggOrganism(record))
        self.records += organisms

        # Build a dictionary to lookup organisms by code.
        for index in range(len(self.records)):
            self.code_to_id[self.records[index].code] = self.records[index].id

        return

    def set_representatives(self, otu_file):
        """ Set the otu_representative flag for organisms from source file.

        The OTU file lists the organisms that are members of an OTU and which organism
        is the OTU representative. Each line in the file has four fields: (1) OTU
        number, (2) boolean flag that identifies OTU representative ('1' means true),
        (3) genome ID, and (4) scientific name of organism.

        A DataFrame is returned with a report that gives details on the analysis done
        to pick the OTU representative organisms. Each row in the DataFrame contains
        the following columns: (1) 'OTU_REP_NAME' name of organism that is OTU representative
        in source file, (2) 'OTU_REP_ID' ID of organism that is OTU representative in source
        file, (3) 'NAME' name of organism that is a possible match, (4) 'ID' ID of organism
        that is a possible match, (5) 'RATIO' value of match ratio, (6) 'PARTIAL_RATIO'
        value of partial match ratio, (7) 'TOKEN_SET_RATIO' value of token set match ratio,
        (8) 'RESULT' result of analysis where 'selected' means the organism was selected as
        the OTU representative, 'alternate' means the organism was a good match but was not
        selected, 'skipped' means the organism was not selected, 'not found' means a good
        match was not found, and 'no match' means the organism was not considered.

        Note that if multiple organisms are equally good matches, it is random which one
        is picked as the representative.

        Parameters
        ----------
        otu_file : str
            Path to file that identifies OTU representatives in TSV format

        Returns
        -------
        pandas.DataFrame
            Report with results of analysis
        """

        if len(self.records) == 0:
            raise DatabaseError('Organism database is empty')

        # Reset all organisms to not be a OTU representative.
        for organism in self.records:
            organism.otu_representative = 0

        # Build a dictionary where the key is name of OTU representative organism and
        # value is organism ID.
        representatives = dict()
        with open(otu_file, 'r') as handle:
            for line in handle:
                fields = line.strip().split('\t')
                if fields[1] == '1':
                    # Remove 'substr' and 'str' from scientific name for better matching to KEGG names.
                    name = fields[3].replace('substr.', '').replace('str.', '')
                    representatives[name] = fields[2]

        # Build a dictionary for finding prokaryote organisms by name where the key
        # is the genus and species portion of the organism name and the value is a
        # list of KeggOrganism objects that match the genus and species. This reduces
        # the number of organisms that need to have a match ratio calculated.
        lookup = dict()
        for organism in self.records:
            if organism.is_prokaryote:
                num_words = 2
                if organism.name.startswith('Candidatus'):
                    num_words = 3  # Need first three words for candidate names
                parts = organism.name.split()
                lookup_name = ' '.join(parts[0:num_words])
                try:
                    lookup[lookup_name].append(organism)
                except KeyError:
                    lookup[lookup_name] = [organism]

        # For each OTU representative from OTU file, see if there is a matching organism
        # from KEGG. If there is a match, mark the organism as a OTU representative.
        report = DataFrame(columns=otu_report_columns)
        for name in representatives:
            num_words = 2
            if name.startswith('Candidatus'):
                num_words = 3  # Need first three words for candidate names
            parts = name.split()
            lookup_name = ' '.join(parts[0:num_words])

            # See if there are potential matches in database.
            if lookup_name in lookup:
                # Calculate the match ratio for all potential matches using the search name.
                matches = list()
                for organism in lookup[lookup_name]:
                    match = dict()
                    match['organism'] = organism
                    match['ratio'] = fuzz.ratio(name, organism.search_name)
                    match['partial_ratio'] = 0
                    match['token_set_ratio'] = 0
                    match['possible'] = False
                    matches.append(match)

                # Sort the matches by the ratio so best matches are first.
                matches = sorted(matches, key=lambda k: k['ratio'], reverse=True)

                # See if there are any good matches and select the best matches.
                possible = list()
                max_ratio = matches[0]['ratio']
                if max_ratio >= MATCH_CUTOFF:
                    for match in matches:
                        if match['ratio'] == max_ratio:
                            possible.append(match)
                            match['possible'] = True

                # If there are no good matches yet, calculate the partial match ratio
                # for all potential matches using the search name.
                if len(possible) == 0:
                    for match in matches:
                        match['partial_ratio'] = fuzz.partial_ratio(name, match['organism'].search_name)
                        if match['partial_ratio'] >= MATCH_CUTOFF:
                            possible.append(match)
                            match['possible'] = True

                # If there are no good matches yet, calculate the token set ratio
                # for all potential matches using the name.
                if len(possible) == 0:
                    for match in matches:
                        match['token_set_ratio'] = fuzz.token_set_ratio(name, match['organism'].name)
                        if match['token_set_ratio'] == 100:
                            possible.append(match)
                            match['possible'] = True

                # Set the OTU representative based on the best match.
                none_selected = False
                if len(possible) > 0:
                    possible[0]['organism'].otu_representative = 1
                else:
                    # There is only possible match that is very close.
                    if len(lookup[lookup_name]) == 1:
                        lookup[lookup_name][0].otu_representative = 1
                    else:
                        none_selected = True

                # Add the results to the report.
                for match in matches:
                    organism = match['organism']
                    if none_selected:
                        selected_value = 'not found'
                    elif len(possible) > 1:
                        if organism.is_representative:
                            selected_value = 'selected'
                        elif match['possible']:
                            selected_value = 'alternate'
                        else:
                            selected_value = 'skipped'
                    else:
                        if organism.is_representative:
                            selected_value = 'selected'
                        else:
                            selected_value = 'skipped'
                    result = Series([name, representatives[name], organism.name, organism.id,
                                    match['ratio'], match['partial_ratio'], match['token_set_ratio'],
                                    selected_value], index=otu_report_columns)
                    report = report.append(result, ignore_index=True)

            else:
                result = Series([name, representatives[name], 'no match', 'no match',
                                0, 0, 0, 'no match'], index=otu_report_columns)
                report = report.append(result, ignore_index=True)

        return report

    def num_otu_representatives(self):
        """ Get the number of OTU representative organisms in the database.

        Returns
        -------
        int
            Number of OTU representative organisms
        """

        return len(self.records.query(lambda x: x, 'is_representative'))

    def num_prokaryotes(self):
        """ Get the number of prokaryote organisms in the database.

        Returns
        -------
        int
            Number of prokaryote organisms
        """

        return len(self.records.query(lambda x: x, 'is_prokaryote'))

    def num_eukaryotes(self):
        """ Get the number of eukaryote organisms in the database.

        Returns
        -------
        int
            Number of eukaryote organisms
        """

        return len(self.records.query(lambda x: x, 'is_eukaryote'))

    def num_bacteria(self):
        """ Get the number of bacteria organisms in the database.

        Returns
        -------
        int
            Number of bacteria organisms
        """

        return len(self.records.query(lambda x: x, 'is_bacteria'))

    def num_archaea(self):
        """ Get the number of archaea organisms in the database.

        Returns
        -------
        int
            Number of archaea organisms
        """

        return len(self.records.query(lambda x: x, 'is_archaea'))

    def has_code(self, code):
        """ Check if a code exists in the organism database.

        Parameters
        ----------
        code : str
            Organism code to check

        Returns
        -------
        bool
            True when code exists, otherwise False
        """

        return code in self.code_to_id

    def get_by_code(self, code):
        """ Get an organism with the specified code.
        
        Parameters
        ----------
        code : str
            Code of record to return

        Returns
        -------
        KeggOrganism
            Organism object with specified code
        """

        return self.records.get_by_id(self.code_to_id[code])
